import asyncio
from typing import Any, Dict, Optional, List
from pydantic import BaseModel

from opal_common.schemas.policy import PolicyBundle
from opal_common.schemas.store import StoreTransaction

from .base_policy_store_client import  BasePolicyStoreClient, JsonableValue

class MockPolicyStoreClient(BasePolicyStoreClient):
    """
    A naive mock policy and policy-data store for tests
    """

    def __init__(self) -> None:
        super().__init__()
        self._has_data_event:asyncio.Event() = None
        self._data = {}

    @property
    def has_data_event(self):
        if self._has_data_event is None:
            self._has_data_event = asyncio.Event()
        return self._has_data_event

    async def set_policy(self, policy_id: str, policy_code: str, transaction_id:Optional[str]=None):
        pass

    async def get_policy(self, policy_id: str) -> Optional[str]:
        pass

    async def delete_policy(self, policy_id: str, transaction_id:Optional[str]=None):
        pass

    async def get_policy_module_ids(self) -> List[str]:
        pass

    async def set_policies(self, bundle: PolicyBundle, transaction_id:Optional[str]=None):
        pass

    async def get_policy_version(self) -> Optional[str]:
        return None

    async def set_policy_data(self, policy_data: JsonableValue, path: str = "", transaction_id:Optional[str]=None):
        self._data[path] = policy_data
        self.has_data_event.set()

    async def get_data(self, path: str=None) -> Dict:
        if path is None or path == "":
            return self._data
        else:
            return self._data[path]

    async def get_data_with_input(self, path: str, input: BaseModel) -> Dict:
        return {}

    async def delete_policy_data(self, path: str = "", transaction_id:Optional[str]=None):
        if not path:
            self._data = {}
        else:
            del self._data[path]

    async def patch_data(self, path: str, patch_document: Dict[str, Any], transaction_id:Optional[str]=None):
        pass

    async def wait_for_data(self):
        """
        Wait until the store has data set in it
        """
        await self.has_data_event.wait()

    async def init_healthcheck_policy(self, policy_id: str, policy_code: str, data_updater_enabled: bool = True):
        pass

    async def persist_transaction(self, transaction: StoreTransaction):
        pass


from enum import Enum
from typing import Optional
from pydantic import BaseModel, Field, validator


class PolicyStoreTypes(Enum):
    OPA = "OPA"
    MOCK = "MOCK"


class PolicyStoreDetails(BaseModel):
    """
    represents a policy store endpoint - contains the policy store's:
    - location (url)
    - type
    - credentials
    """
    type: PolicyStoreTypes = Field(PolicyStoreTypes.OPA, description="the type of policy store, currently only OPA is officially supported")
    url: str = Field(..., description="the url that OPA can be found in. if localhost is the host - it means OPA is on the same hostname as OPAL client.")
    token: Optional[str] = Field(None, description="optional access token required by the policy store")

    @validator('type')
    def force_enum(cls, v):
        if isinstance(v, str):
            return PolicyStoreTypes(v)
        if isinstance(v, PolicyStoreTypes):
            return v
        raise ValueError(f'invalid value: {v}')

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True



from logging import disable
import os
import signal
import asyncio
import uuid
import aiohttp
import functools
from typing import List, Optional

from fastapi import FastAPI
import websockets

from opal_common.logger import logger, configure_logs
from opal_common.middleware import configure_middleware
from opal_common.config import opal_common_config
from opal_common.security.sslcontext import get_custom_ssl_context
from opal_common.authentication.verifier import JWTVerifier
from opal_common.authentication.deps import JWTAuthenticator
from opal_client.policy_store.api import init_policy_store_router
from opal_client.config import PolicyStoreTypes, opal_client_config
from opal_client.data.api import init_data_router
from opal_client.data.updater import DataUpdater
from opal_client.data.fetcher import DataFetcher
from opal_client.policy_store.base_policy_store_client import BasePolicyStoreClient
from opal_client.policy_store.policy_store_client_factory import PolicyStoreClientFactory
from opal_client.opa.runner import OpaRunner
from opal_client.opa.options import OpaServerOptions
from opal_client.policy.api import init_policy_router
from opal_client.policy.updater import PolicyUpdater
from opal_client.callbacks.register import CallbacksRegister
from opal_client.callbacks.api import init_callbacks_api


class OpalClient:
    def __init__(
        self,
        policy_store_type:PolicyStoreTypes=None,
        policy_store:BasePolicyStoreClient=None,
        data_updater:DataUpdater=None,
        data_topics: List[str] = None,
        policy_updater:PolicyUpdater=None,
        inline_opa_enabled:bool=None,
        inline_opa_options:OpaServerOptions=None,
        verifier: Optional[JWTVerifier] = None,
    ) -> None:
        """
        Args:
            policy_store_type (PolicyStoreTypes, optional): [description]. Defaults to POLICY_STORE_TYPE.

            Internal components (for each pass None for default init, or False to disable):
                policy_store (BasePolicyStoreClient, optional): The policy store client. Defaults to None.
                data_updater (DataUpdater, optional): Defaults to None.
                policy_updater (PolicyUpdater, optional): Defaults to None.
        """
        # defaults
        policy_store_type: PolicyStoreTypes = policy_store_type or opal_client_config.POLICY_STORE_TYPE
        inline_opa_enabled: bool = inline_opa_enabled or opal_client_config.INLINE_OPA_ENABLED
        inline_opa_options: OpaServerOptions = inline_opa_options or opal_client_config.INLINE_OPA_CONFIG
        opal_client_identifier: str = opal_client_config.OPAL_CLIENT_STAT_ID or f"CLIENT_{uuid.uuid4().hex}"
        # set logs
        configure_logs()
        # Init policy store client
        self.policy_store_type: PolicyStoreTypes = policy_store_type
        self.policy_store: BasePolicyStoreClient = policy_store or PolicyStoreClientFactory.create(policy_store_type)
        # data fetcher
        self.data_fetcher = DataFetcher()
        # callbacks register
        if hasattr(opal_client_config.DEFAULT_UPDATE_CALLBACKS, 'callbacks'):
            default_callbacks = opal_client_config.DEFAULT_UPDATE_CALLBACKS.callbacks
        else:
            default_callbacks = []

        self._callbacks_register = CallbacksRegister(default_callbacks)

        # Init policy updater
        if policy_updater is not None:
            self.policy_updater = policy_updater
        else:
            self.policy_updater = PolicyUpdater(policy_store=self.policy_store, data_fetcher=self.data_fetcher, callbacks_register=self._callbacks_register, opal_client_id=opal_client_identifier)
        # Data updating service
        if opal_client_config.DATA_UPDATER_ENABLED:
            if data_updater is not None:
                self.data_updater = data_updater
            else:
                data_topics = data_topics if data_topics is not None else opal_client_config.DATA_TOPICS
                self.data_updater = DataUpdater(policy_store=self.policy_store, data_topics=data_topics, data_fetcher=self.data_fetcher, callbacks_register=self._callbacks_register, opal_client_id=opal_client_identifier)
        else:
            self.data_updater = None

        # Internal services
        # Policy store
        if self.policy_store_type == PolicyStoreTypes.OPA and inline_opa_enabled:
            rehydration_callbacks = [
                # refetches policy code (e.g: rego) and static data from server
                functools.partial(self.policy_updater.update_policy, force_full_update=True),
            ]

            if self.data_updater:
                rehydration_callbacks.append(
                    functools.partial(self.data_updater.get_base_policy_data, data_fetch_reason="policy store rehydration")
                )

            self.opa_runner = OpaRunner.setup_opa_runner(options=inline_opa_options, rehydration_callbacks=rehydration_callbacks)
        else:
            self.opa_runner = False


        custom_ssl_context = get_custom_ssl_context()
        if opal_common_config.CLIENT_SELF_SIGNED_CERTIFICATES_ALLOWED and custom_ssl_context is not None:
            logger.warning("OPAL client is configured to trust self-signed certificates")

        if verifier is not None:
            self.verifier = verifier
        else:
            self.verifier = JWTVerifier(
                public_key=opal_common_config.AUTH_PUBLIC_KEY,
                algorithm=opal_common_config.AUTH_JWT_ALGORITHM,
                audience=opal_common_config.AUTH_JWT_AUDIENCE,
                issuer=opal_common_config.AUTH_JWT_ISSUER,
            )
        if not self.verifier.enabled:
            logger.info("API authentication disabled (public encryption key was not provided)")

        # init fastapi app
        self.app: FastAPI = self._init_fast_api_app()

    def _init_fast_api_app(self):
        """
        inits the fastapi app object
        """
        app = FastAPI(
            title="OPAL Client",
            description="OPAL is an administration layer for Open Policy Agent (OPA), detecting changes" + \
            " to both policy and data and pushing live updates to your agents. The opal client is" + \
            " deployed alongside a policy-store (e.g: OPA), keeping it up-to-date, by connecting to" + \
            " an opal-server and subscribing to pub/sub updates for policy and policy data changes.",
            version="0.1.0"
        )
        configure_middleware(app)
        self._configure_api_routes(app)
        self._configure_lifecycle_callbacks(app)
        return app

    def _configure_api_routes(self, app: FastAPI):
        """
        mounts the api routes on the app object
        """

        authenticator = JWTAuthenticator(self.verifier)

        # Init api routers with required dependencies
        policy_router = init_policy_router(policy_updater=self.policy_updater)
        data_router = init_data_router(data_updater=self.data_updater)
        policy_store_router = init_policy_store_router(authenticator)
        callbacks_router = init_callbacks_api(authenticator, self._callbacks_register)

        # mount the api routes on the app object
        app.include_router(policy_router, tags=["Policy Updater"])
        app.include_router(data_router, tags=["Data Updater"])
        app.include_router(policy_store_router, tags=["Policy Store"])
        app.include_router(callbacks_router, tags=["Callbacks"])

        # top level routes (i.e: healthchecks)
        @app.get("/healthcheck", include_in_schema=False)
        @app.get("/", include_in_schema=False)
        def healthcheck():
            return {"status": "ok"}
        return app

    def _configure_lifecycle_callbacks(self, app: FastAPI):
        """
        registers callbacks on app startup and shutdown.

        on app startup we launch our long running processes (async tasks)
        on the event loop. on app shutdown we stop these long running tasks.
        """
        @app.on_event("startup")
        async def startup_event():
            asyncio.create_task(self.start_client_background_tasks())

        @app.on_event("shutdown")
        async def shutdown_event():
            await self.stop_client_background_tasks()

        return app

    async def start_client_background_tasks(self):
        """
        Launch OPAL client long-running tasks:
        - Policy Store runner (e.g: Opa Runner)
        - Policy Updater
        - Data Updater

        If there is a policy store to run, we wait until its up before launching dependent tasks.
        """
        if self.opa_runner:
            # runs the policy store dependent tasks after policy store is up
            self.opa_runner.register_opa_initial_start_callbacks([self.launch_policy_store_dependent_tasks])
            async with self.opa_runner:
                await self.opa_runner.wait_until_done()
        else:
            # we do not run the policy store in the same container
            # therefore we can immediately launch dependent tasks
            await self.launch_policy_store_dependent_tasks()

    async def stop_client_background_tasks(self):
        """
        stops all background tasks (called on shutdown event)
        """
        logger.info("stopping background tasks...")

        # stopping opa runner
        if self.opa_runner:
            await self.opa_runner.stop()

        # stopping updater tasks (each updater runs a pub/sub client)
        logger.info("trying to shutdown DataUpdater and PolicyUpdater gracefully...")
        tasks: List[asyncio.Task] = []
        if self.data_updater:
            tasks.append(asyncio.create_task(self.data_updater.stop()))
        if self.policy_updater:
            tasks.append(asyncio.create_task(self.policy_updater.stop()))

        try:
            await asyncio.gather(*tasks)
        except Exception:
            logger.exception("exception while shutting down updaters")

    async def launch_policy_store_dependent_tasks(self):
        try:
            await self.maybe_init_healthcheck_policy()
        except Exception:
            logger.critical("healthcheck policy enabled but could not be initialized!")
            self._trigger_shutdown()
            return

        try:
            for task in asyncio.as_completed([self.launch_policy_updater(), self.launch_data_updater()]):
                await task
        except websockets.exceptions.InvalidStatusCode as err:
            logger.error("Failed to launch background task -- {err}", err=repr(err))
            self._trigger_shutdown()

    async def maybe_init_healthcheck_policy(self):
        """
        This function only runs if OPA_HEALTH_CHECK_POLICY_ENABLED is true.

        Puts the healthcheck policy in opa cache and inits the transaction log used by the policy.
        If any action fails, opal client will shutdown.
        """
        if not opal_client_config.OPA_HEALTH_CHECK_POLICY_ENABLED:
            return # skip

        healthcheck_policy_relpath = opal_client_config.OPA_HEALTH_CHECK_POLICY_PATH

        here = os.path.abspath(os.path.dirname(__file__))
        healthcheck_policy_path = os.path.join(here, healthcheck_policy_relpath)
        if not os.path.exists(healthcheck_policy_path):
            logger.error("Critical: OPA health-check policy is enabled, but cannot find policy at {path}", path=healthcheck_policy_path)
            raise ValueError("OPA health check policy not found!")

        try:
            healthcheck_policy_code = open(healthcheck_policy_path, 'r').read()
        except IOError as err:
            logger.error("Critical: Cannot read healthcheck policy: {err}", err=repr(err))
            raise

        try:
            await self.policy_store.init_healthcheck_policy(
                policy_id=healthcheck_policy_relpath,
                policy_code=healthcheck_policy_code,
                data_updater_enabled=opal_client_config.DATA_UPDATER_ENABLED
            )
        except aiohttp.ClientError as err:
            logger.error("Failed to connect to OPA agent while init healthcheck policy -- {err}", err=repr(err))
            raise

    def _trigger_shutdown(self):
        """
        this will send SIGTERM (Keyboard interrupt) to the worker, making uvicorn
        send "lifespan.shutdown" event to Starlette via the ASGI lifespan interface.
        Starlette will then trigger the @app.on_event("shutdown") callback, which
        in our case (self.stop_client_background_tasks()) will gracefully shutdown
        the background processes and only then will terminate the worker.
        """
        logger.info("triggering shutdown with SIGTERM...")
        os.kill(os.getpid(), signal.SIGTERM)

    async def launch_policy_updater(self):
        if self.policy_updater:
            async with self.policy_updater:
                await self.policy_updater.wait_until_done()

    async def launch_data_updater(self):
        if self.data_updater:
            async with self.data_updater:
                await self.data_updater.wait_until_done()


import asyncio
from opal_common.schemas.data import DataUpdateReport
from opal_client.callbacks.reporter import CallbacksReporter
from opal_client.callbacks.register import CallbacksRegister
from opal_client.data.fetcher import DataFetcher
from typing import List, Optional

import pydantic
from fastapi_websocket_rpc.rpc_channel import RpcChannel
from fastapi_websocket_pubsub import PubSubClient
from opal_common.schemas.store import TransactionType

from opal_common.config import opal_common_config
from opal_common.utils import get_authorization_header
from opal_common.schemas.policy import PolicyBundle, PolicyUpdateMessage
from opal_common.security.sslcontext import get_custom_ssl_context
from opal_common.topics.utils import pubsub_topics_from_directories
from opal_client.logger import logger
from opal_client.config import opal_client_config
from opal_client.policy.fetcher import PolicyFetcher
from opal_client.policy_store.base_policy_store_client import BasePolicyStoreClient
from opal_client.policy_store.policy_store_client_factory import DEFAULT_POLICY_STORE_GETTER
from opal_client.policy.topics import default_subscribed_policy_directories

class PolicyUpdater:
    """
    Keeps policy-stores (e.g. OPA) up to date with relevant policy code
    (e.g: rego) and static data (e.g: data.json files like in OPA bundles).

    Uses Pub/Sub to subscribe to specific directories in the policy code
    repository (i.e: git), and fetches bundles containing updated policy code.
    """
    def __init__(
        self,
        token: str = None,
        pubsub_url: str = None,
        subscription_directories: List[str] = None,
        policy_store: BasePolicyStoreClient = None,
        data_fetcher: Optional[DataFetcher] = None,
        callbacks_register: Optional[CallbacksRegister] = None,
        opal_client_id: str = None
    ):
        """inits the policy updater.

        Args:
            token (str, optional): Auth token to include in connections to OPAL server. Defaults to CLIENT_TOKEN.
            pubsub_url (str, optional): URL for Pub/Sub updates for policy. Defaults to OPAL_SERVER_PUBSUB_URL.
            subscription_directories (List[str], optional): directories in the policy source repo to subscribe to.
                Defaults to POLICY_SUBSCRIPTION_DIRS. every time the directory is updated by a commit we will receive
                a message on its respective topic. we dedups directories with ancestral relation, and will only
                receive one message for each updated file.
            policy_store (BasePolicyStoreClient, optional): Policy store client to use to store policy code. Defaults to DEFAULT_POLICY_STORE.
        """
        # defaults
        token: str = token or opal_client_config.CLIENT_TOKEN
        pubsub_url: str = pubsub_url or opal_client_config.SERVER_PUBSUB_URL
        self._subscription_directories: List[str] = subscription_directories or opal_client_config.POLICY_SUBSCRIPTION_DIRS
        self._opal_client_id = opal_client_id

        # The policy store we'll save policy modules into (i.e: OPA)
        self._policy_store = policy_store or DEFAULT_POLICY_STORE_GETTER()
        # pub/sub server url and authentication data
        self._server_url = pubsub_url
        self._token = token
        if self._token is None:
            self._extra_headers = None
        else:
            self._extra_headers = [get_authorization_header(self._token)]
        # Pub/Sub topics we subscribe to for policy updates
        self._topics = pubsub_topics_from_directories(self._subscription_directories)
        # The pub/sub client for data updates
        self._client = None
        # The task running the Pub/Sub subcribing client
        self._subscriber_task = None
        self._stopping = False
        # policy fetcher - fetches policy bundles
        self._policy_fetcher = PolicyFetcher()
        # callbacks on policy changes
        self._data_fetcher = data_fetcher or DataFetcher()
        self._callbacks_register = callbacks_register or CallbacksRegister()
        self._callbacks_reporter = CallbacksReporter(self._callbacks_register, self._data_fetcher)
        self._should_send_reports = opal_client_config.SHOULD_REPORT_ON_DATA_UPDATES or False
        # custom SSL context (for self-signed certificates)
        self._custom_ssl_context = get_custom_ssl_context()
        self._ssl_context_kwargs = {'ssl': self._custom_ssl_context} if self._custom_ssl_context is not None else {}

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if not self._stopping:
            await self.stop()

    async def _update_policy_callback(self, data: dict = None, topic: str = "", **kwargs):
        """
        Pub/Sub callback - triggering policy updates
        will run when we get notifications on the policy topic.
        i.e: when the source repository changes (new commits)
        """
        if data is None:
            logger.warning("got policy update message without data, skipping policy update!")
            return

        try:
            message = PolicyUpdateMessage(**data)
        except pydantic.ValidationError as e:
            logger.warning(f"Got invalid policy update message from server: {repr(e)}")
            return

        logger.info(
            "Received policy update: topic={topic}, message={message}",
            topic=topic,
            message=message.dict()
        )

        directories = list(set(message.changed_directories).intersection(set(self._subscription_directories)))
        await self.update_policy(directories)

    async def _on_connect(self, client: PubSubClient, channel: RpcChannel):
        """
        Pub/Sub on_connect callback
        On connection to backend, whether its the first connection,
        or reconnecting after downtime, refetch the state opa needs.
        As long as the connection is alive we know we are in sync with the server,
        when the connection is lost we assume we need to start from scratch.
        """
        logger.info("Connected to server")
        await self.update_policy()
        if opal_common_config.STATISTICS_ENABLED:
            await self._client.wait_until_ready()
            # publish statistics to the server about new connection from client (only if STATISTICS_ENABLED is True, default to False)
            await self._client.publish([opal_common_config.STATISTICS_ADD_CLIENT_CHANNEL], data={'topics': self._topics, 'client_id': self._opal_client_id, 'rpc_id': channel.id})

    async def _on_disconnect(self, channel: RpcChannel):
        """
        Pub/Sub on_disconnect callback
        """
        logger.info("Disconnected from server")

    async def start(self):
        """
        launches the policy updater
        """
        logger.info("Launching policy updater")
        if self._subscriber_task is None:
            self._subscriber_task = asyncio.create_task(self._subscriber())
            await self._data_fetcher.start()

    async def stop(self):
        """
        stops the policy updater
        """
        self._stopping = True
        logger.info("Stopping policy updater")

        # disconnect from Pub/Sub
        if self._client is not None:
            try:
                await asyncio.wait_for(self._client.disconnect(), timeout=3)
            except asyncio.TimeoutError:
                logger.debug("Timeout waiting for PolicyUpdater pubsub client to disconnect")

        # stop subscriber task
        if self._subscriber_task is not None:
            logger.debug("Cancelling PolicyUpdater subscriber task")
            self._subscriber_task.cancel()
            try:
                await self._subscriber_task
            except asyncio.CancelledError as exc:
                logger.debug("PolicyUpdater subscriber task was force-cancelled: {exc}", exc=repr(exc))
            self._subscriber_task = None
            logger.debug("PolicyUpdater subscriber task was cancelled")

        await self._data_fetcher.stop()

    async def wait_until_done(self):
        if self._subscriber_task is not None:
            await self._subscriber_task

    async def _subscriber(self):
        """
        Coroutine meant to be spunoff with create_task to listen in
        the background for policy update events and pass them to the
        update_policy() callback (which will fetch the relevant policy
        bundle from the server and update the policy store).
        """
        logger.info("Subscribing to topics: {topics}", topics=self._topics)
        self._client = PubSubClient(
            topics=self._topics,
            callback=self._update_policy_callback,
            on_connect=[self._on_connect],
            on_disconnect=[self._on_disconnect],
            extra_headers=self._extra_headers,
            keep_alive=opal_client_config.KEEP_ALIVE_INTERVAL,
            server_uri=self._server_url,
            **self._ssl_context_kwargs
        )
        async with self._client:
            await self._client.wait_until_done()

    async def update_policy(self, directories: List[str] = None, force_full_update=False):
        """
        fetches policy (code, e.g: rego) from backend and stores it in the policy store.

        Args:
            policy_store (BasePolicyStoreClient, optional): Policy store client to use to store policy code.
            directories (List[str], optional): specific source directories we want.
            force_full_update (bool, optional): if true, ignore stored hash and fetch full policy bundle.
        """
        directories = directories if directories is not None else default_subscribed_policy_directories()
        if force_full_update:
            logger.info("full update was forced (ignoring stored hash if exists)")
            base_hash = None
        else:
            base_hash = await self._policy_store.get_policy_version()

        if base_hash is None:
            logger.info("Refetching policy code (full bundle)")
        else:
            logger.info("Refetching policy code (delta bundle), base hash: '{base_hash}'", base_hash=base_hash)
        bundle_error = None
        bundle = None
        bundle_succeeded = True
        try:
            bundle: Optional[PolicyBundle] = await self._policy_fetcher.fetch_policy_bundle(directories, base_hash=base_hash)
            if bundle:
                if bundle.old_hash is None:
                    logger.info(
                    "Got policy bundle with {rego_files} rego files, {data_files} data files, commit hash: '{commit_hash}'",
                    rego_files=len(bundle.policy_modules),
                    data_files=len(bundle.data_modules),
                        commit_hash=bundle.hash,
                        manifest=bundle.manifest
                    )
                else:
                    deleted_files = None if bundle.deleted_files is None else bundle.deleted_files.dict()
                    logger.info(
                        "got policy bundle (delta): '{diff_against_hash}' -> '{commit_hash}'",
                        commit_hash=bundle.hash,
                        diff_against_hash=bundle.old_hash,
                        manifest=bundle.manifest,
                        deleted=deleted_files
                    )
        except Exception as err:
            bundle_error = repr(err)
            bundle_succeeded = False
        bundle_hash = None if bundle is None else bundle.hash

        # store policy bundle in OPA cache
        # We wrap our interaction with the policy store with a transaction, so that
        # if the write-op fails, we will mark the transaction as failed.
        async with self._policy_store.transaction_context(bundle_hash, transaction_type=TransactionType.policy) as store_transaction:
            store_transaction._update_remote_status(url=self._policy_fetcher.policy_endpoint_url, status=bundle_succeeded, error=bundle_error)
            if bundle:
                await store_transaction.set_policies(bundle)
                # if we got here, we did not throw during the transaction
                if self._should_send_reports:
                    # spin off reporting (no need to wait on it)
                    report = DataUpdateReport(policy_hash=bundle.hash, reports=[])
                    asyncio.create_task(self._callbacks_reporter.report_update_results(report))




import os
import sys
from fastapi.applications import FastAPI
import typer
from typer.main import Typer
from typer.models import Context

# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
sys.path.append(root_dir)

from opal_client.config import opal_client_config
from opal_common.config import opal_common_config
from opal_common.corn_utils import run_gunicorn, run_uvicorn
from opal_common.cli.typer_app import get_typer_app
from opal_common.cli.docs import MainTexts

app = get_typer_app()


@app.command()
def run(engine_type: str = typer.Option("uvicron", help="uvicorn or gunicorn")):
    """
    Run the client as a deamon 
    """
    typer.echo(f"-- Starting OPAL client (with {engine_type}) --")

    if engine_type == 'gunicorn':
        app:FastAPI 
        from opal_client.main import app
        run_gunicorn(app, opal_client_config.CLIENT_API_SERVER_WORKER_COUNT,
                     host=opal_client_config.CLIENT_API_SERVER_HOST,
                     port=opal_client_config.CLIENT_API_SERVER_PORT)
    else:
        run_uvicorn("opal_client.main:app",
                    workers=opal_client_config.CLIENT_API_SERVER_WORKER_COUNT,
                    host=opal_client_config.CLIENT_API_SERVER_HOST,
                    port=opal_client_config.CLIENT_API_SERVER_PORT)


@app.command()
def print_config():
    """
    To test config values, print the configuration parsed from ENV and CMD
    """
    typer.echo("Printing configuration values")
    typer.echo(str(opal_client_config))
    typer.echo(str(opal_common_config))




def cli():
    main_texts = MainTexts("OPAL-CLIENT", "client")
    def on_start(ctx:Context, **kwargs):
        if ctx.invoked_subcommand is None or ctx.invoked_subcommand == "run":
            typer.secho(main_texts.header, bold=True, fg=typer.colors.MAGENTA)
        if ctx.invoked_subcommand is None:
            typer.echo(ctx.get_usage())
            typer.echo(main_texts.docs)
            
    opal_client_config.cli([opal_common_config], typer_app=app, help=main_texts.docs, on_start=on_start)

if __name__ == "__main__":
    cli()



from fastapi import APIRouter, status

from opal_common.logger import logger
from opal_client.policy.updater import PolicyUpdater

def init_policy_router(policy_updater: PolicyUpdater):
    router = APIRouter()

    @router.post("/policy-updater/trigger", status_code=status.HTTP_200_OK)
    async def trigger_policy_update():
        logger.info("triggered policy update from api")
        await policy_updater.update_policy(force_full_update=True)
        return {"status": "ok"}

    return router



import asyncio
import hashlib
import itertools
import json
from typing import List, Optional, Tuple
import uuid
import aiohttp

from aiohttp.client import ClientSession, ClientError
from fastapi_websocket_pubsub import PubSubClient
from fastapi_websocket_rpc.rpc_channel import RpcChannel

from opal_common.config import opal_common_config
from opal_common.fetcher.events import FetcherConfig
from opal_common.schemas.data import (DataEntryReport, DataSourceConfig,
                                      DataSourceEntry, DataUpdate,
                                      DataUpdateReport)
from opal_common.schemas.store import TransactionType
from opal_common.utils import get_authorization_header
from opal_common.http import is_http_error_response
from opal_common.security.sslcontext import get_custom_ssl_context
from opal_client.callbacks.reporter import CallbacksReporter
from opal_client.callbacks.register import CallbacksRegister
from opal_client.config import opal_client_config
from opal_client.data.fetcher import DataFetcher
from opal_client.data.rpc import TenantAwareRpcEventClientMethods
from opal_client.logger import logger
from opal_client.policy_store.base_policy_store_client import \
    BasePolicyStoreClient
from opal_client.policy_store.policy_store_client_factory import \
    DEFAULT_POLICY_STORE_GETTER


class DataUpdater:
    def __init__(self, token: str = None,
                 pubsub_url: str = None,
                 data_sources_config_url: str = None,
                 fetch_on_connect: bool = True,
                 data_topics: List[str] = None,
                 policy_store: BasePolicyStoreClient = None,
                 should_send_reports=None,
                 data_fetcher: Optional[DataFetcher] = None,
                 callbacks_register: Optional[CallbacksRegister] = None,
                 opal_client_id: str = None
        ):
        """
        Keeps policy-stores (e.g. OPA) up to date with relevant data
        Obtains data configuration on startup from OPAL-server
        Uses Pub/Sub to subscribe to data update events, and fetches (using FetchingEngine) data from sources.

        Args:
            token (str, optional): Auth token to include in connections to OPAL server. Defaults to CLIENT_TOKEN.
            pubsub_url (str, optional): URL for Pub/Sub updates for data. Defaults to OPAL_SERVER_PUBSUB_URL.
            data_sources_config_url (str, optional): URL to retrive base data configuration. Defaults to DEFAULT_DATA_SOURCES_CONFIG_URL.
            fetch_on_connect (bool, optional): Should the update fetch basic data immediately upon connection/reconnection. Defaults to True.
            data_topics (List[str], optional): Topics of data to fetch and subscribe to. Defaults to DATA_TOPICS.
            policy_store (BasePolicyStoreClient, optional): Policy store client to use to store data. Defaults to DEFAULT_POLICY_STORE.
        """
        # Defaults
        token: str = token or opal_client_config.CLIENT_TOKEN
        pubsub_url: str = pubsub_url or opal_client_config.SERVER_PUBSUB_URL
        data_sources_config_url: str = data_sources_config_url or opal_client_config.DEFAULT_DATA_SOURCES_CONFIG_URL
        # Should the client use the default data source to fetch on connect
        self._fetch_on_connect = fetch_on_connect
        # The policy store we'll save data updates into
        self._policy_store = policy_store or DEFAULT_POLICY_STORE_GETTER()
        # Pub/Sub topics we subscribe to for data updates
        self._data_topics = data_topics if data_topics is not None else opal_client_config.DATA_TOPICS
        self._should_send_reports = should_send_reports if should_send_reports is not None else opal_client_config.SHOULD_REPORT_ON_DATA_UPDATES
        # The pub/sub client for data updates
        self._client = None
        # The task running the Pub/Sub subcribing client
        self._subscriber_task = None
        # Data fetcher
        self._data_fetcher = data_fetcher or DataFetcher()
        self._callbacks_register = callbacks_register or CallbacksRegister()
        self._callbacks_reporter = CallbacksReporter(self._callbacks_register, self._data_fetcher)
        self._token = token
        self._server_url = pubsub_url
        self._data_sources_config_url = data_sources_config_url
        self._opal_client_id = opal_client_id
        if self._token is None:
            self._extra_headers = None
        else:
            self._extra_headers = [get_authorization_header(self._token)]
        self._stopping = False
        # custom SSL context (for self-signed certificates)
        self._custom_ssl_context = get_custom_ssl_context()
        self._ssl_context_kwargs = {'ssl': self._custom_ssl_context} if self._custom_ssl_context is not None else {}

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        Context handler to terminate internal tasks
        """
        if not self._stopping:
            await self.stop()

    async def _update_policy_data_callback(self, data: dict = None, topic=""):
        """
        Pub/Sub callback - triggering data updates
        will run when we get notifications on the policy_data topic.
        i.e: when new roles are added, changes to permissions, etc.
        """
        if data is not None:
            reason = data.get("reason", "")
        else:
            reason = "Periodic update"
        logger.info("Updating policy data, reason: {reason}", reason=reason)
        update = DataUpdate.parse_obj(data)
        self.trigger_data_update(update)

    def trigger_data_update(self, update: DataUpdate):
        # make sure the id has a unique id for tracking
        if update.id is None:
            update.id = uuid.uuid4().hex
        logger.info("Triggering data update with id: {id}", id=update.id)
        asyncio.create_task(self.update_policy_data(
            update, policy_store=self._policy_store, data_fetcher=self._data_fetcher))

    async def get_policy_data_config(self, url: str = None) -> DataSourceConfig:
        """
        Get the configuration for
        Args:
            url: the URL to query for the config, Defaults to self._data_sources_config_url
        Returns:
            DataSourceConfig: the data sources config
        """
        if url is None:
            url = self._data_sources_config_url
        logger.info("Getting data-sources configuration from '{source}'", source=url)
        try:
            async with ClientSession(headers=self._extra_headers) as session:
                response = await session.get(url, **self._ssl_context_kwargs)
                if response.status == 200:
                    return DataSourceConfig.parse_obj(await response.json())
                else:
                    error_details = await response.json()
                    raise ClientError(f"Fetch data sources failed with status code {response.status}, error: {error_details}")
        except:
            logger.exception(f"Failed to load data sources config")
            raise

    async def get_base_policy_data(self, config_url: str = None, data_fetch_reason="Initial load"):
        """
        Load data into the policy store according to the data source's config provided in the config URL

        Args:
            config_url (str, optional): URL to retrive data sources config from. Defaults to None ( self._data_sources_config_url).
            data_fetch_reason (str, optional): Reason to log for the update operation. Defaults to "Initial load".
        """
        logger.info("Performing data configuration, reason: {reason}", reason=data_fetch_reason)
        sources_config = await self.get_policy_data_config(url=config_url)
        # translate config to a data update
        entries = sources_config.entries
        update = DataUpdate(reason=data_fetch_reason, entries=entries)
        self.trigger_data_update(update)

    async def on_connect(self, client: PubSubClient, channel: RpcChannel):
        """
        Pub/Sub on_connect callback
        On connection to backend, whether its the first connection,
        or reconnecting after downtime, refetch the state opa needs.
        As long as the connection is alive we know we are in sync with the server,
        when the connection is lost we assume we need to start from scratch.
        """
        logger.info("Connected to server")
        if self._fetch_on_connect:
            await self.get_base_policy_data()
        if opal_common_config.STATISTICS_ENABLED:
            await self._client.wait_until_ready()
            # publish statistics to the server about new connection from client (only if STATISTICS_ENABLED is True, default to False)
            await self._client.publish([opal_common_config.STATISTICS_ADD_CLIENT_CHANNEL], data={'topics': self._data_topics, 'client_id': self._opal_client_id, 'rpc_id': channel.id})


    async def on_disconnect(self, channel: RpcChannel):
        logger.info("Disconnected from server")

    async def start(self):
        logger.info("Launching data updater")
        if self._subscriber_task is None:
            self._subscriber_task = asyncio.create_task(self._subscriber())
            await self._data_fetcher.start()

    async def _subscriber(self):
        """
        Coroutine meant to be spunoff with create_task to listen in
        the background for data events and pass them to the data_fetcher
        """
        logger.info("Subscribing to topics: {topics}", topics=self._data_topics)
        self._client = PubSubClient(
            self._data_topics,
            self._update_policy_data_callback,
            methods_class=TenantAwareRpcEventClientMethods,
            on_connect=[self.on_connect],
            extra_headers=self._extra_headers,
            keep_alive=opal_client_config.KEEP_ALIVE_INTERVAL,
            server_uri=self._server_url,
            **self._ssl_context_kwargs
        )
        async with self._client:
            await self._client.wait_until_done()

    async def stop(self):
        self._stopping = True
        logger.info("Stopping data updater")

        # disconnect from Pub/Sub
        if self._client is not None:
            try:
                await asyncio.wait_for(self._client.disconnect(), timeout=3)
            except asyncio.TimeoutError:
                logger.debug("Timeout waiting for DataUpdater pubsub client to disconnect")

        # stop subscriber task
        if self._subscriber_task is not None:
            logger.debug("Cancelling DataUpdater subscriber task")
            self._subscriber_task.cancel()
            try:
                await self._subscriber_task
            except asyncio.CancelledError as exc:
                logger.debug("DataUpdater subscriber task was force-cancelled: {exc}", exc=repr(exc))
            self._subscriber_task = None
            logger.debug("DataUpdater subscriber task was cancelled")

        # stop the data fetcher
        logger.debug("Stopping data fetcher")
        await self._data_fetcher.stop()

    async def wait_until_done(self):
        if self._subscriber_task is not None:
            await self._subscriber_task

    @staticmethod
    def calc_hash(data):
        """
        Calculate an hash (sah256) on the given data, if data isn't a string, it will be converted to JSON.
        String are encoded as 'utf-8' prior to hash calculation.
        Returns:
            the hash of the given data (as a a hexdigit string) or '' on failure to process.
        """
        try:
            if not isinstance(data, str):
                data = json.dumps(data, default=str)
            return hashlib.sha256(data.encode('utf-8')).hexdigest()
        except:
            logger.exception("Failed to calculate hash for data {data}", data=data)
            return ""

    async def update_policy_data(self, update: DataUpdate = None, policy_store: BasePolicyStoreClient = None, data_fetcher=None):
        """
        fetches policy data (policy configuration) from backend and updates it into policy-store (i.e. OPA)
        """
        policy_store = policy_store or DEFAULT_POLICY_STORE_GETTER()
        if data_fetcher is None:
            data_fetcher = DataFetcher()
        # types / defaults
        urls: List[Tuple[str, FetcherConfig]] = None
        entries: List[DataSourceEntry] = []
        # track the result of each url in order to report back
        reports: List[DataEntryReport] = []
        # if we have an actual specification for the update
        if update is not None:
            entries = update.entries
            urls = [(entry.url, entry.config) for entry in entries]

        # get the data for the update
        logger.info("Fetching policy data", urls=repr(urls))
        # Urls may be None - handle_urls has a default for None
        policy_data_with_urls = await data_fetcher.handle_urls(urls)
        # Save the data from the update
        # We wrap our interaction with the policy store with a transaction
        async with policy_store.transaction_context(update.id, transaction_type=TransactionType.data) as store_transaction:
            # for intelisense treat store_transaction as a PolicyStoreClient (which it proxies)
            store_transaction: BasePolicyStoreClient
            error_content = None
            for (url, fetch_config, result), entry in itertools.zip_longest(policy_data_with_urls, entries):
                fetched_data_successfully = True

                if isinstance(result, Exception):
                    fetched_data_successfully = False
                    logger.error("Failed to fetch url {url}, got exception: {exc}", url=url, exc=result)

                if isinstance(result, aiohttp.ClientResponse) and is_http_error_response(result): # error responses
                    fetched_data_successfully = False
                    try:
                        error_content = await result.json()
                        logger.error(
                            "Failed to fetch url {url}, got response code {status} with error: {error}",
                            url=url,
                            status=result.status,
                            error=error_content
                        )
                    except json.JSONDecodeError:
                        error_content = await result.text()
                        logger.error(
                            "Failed to decode response from url:{url}, got response code {status} with response: {error}",
                            url=url,
                            status=result.status,
                            error=error_content
                        )
                store_transaction._update_remote_status(url=url, status=fetched_data_successfully, error=str(error_content))

                if fetched_data_successfully:
                    # get path to store the URL data (default mode (None) is as "" - i.e. as all the data at root)
                    policy_store_path = "" if entry is None else entry.dst_path
                    # None is not valid - use "" (protect from missconfig)
                    if policy_store_path is None:
                        policy_store_path = ""
                    # fix opa_path (if not empty must start with "/" to be nested under data)
                    if policy_store_path != "" and not policy_store_path.startswith("/"):
                        policy_store_path = f"/{policy_store_path}"
                    policy_data = result
                    # Create a report on the data-fetching
                    report = DataEntryReport(entry=entry, hash=self.calc_hash(policy_data), fetched=True)
                    logger.info(
                        "Saving fetched data to policy-store: source url='{url}', destination path='{path}'",
                        url=url,
                        path=policy_store_path or '/'
                    )
                    try:
                        await store_transaction.set_policy_data(policy_data, path=policy_store_path)
                        # No exception we we're able to save to the policy-store
                        report.saved = True
                        # save the report for the entry
                        reports.append(report)
                    except Exception:
                        logger.exception("Failed to save data update to policy-store")
                        # we failed to save to policy-store
                        report.saved = False
                        # save the report for the entry
                        reports.append(report)
                        # re-raise so the context manager will be aware of the failure
                        raise
                else:
                    report = DataEntryReport(entry=entry, fetched=False, saved=False)
                    # save the report for the entry
                    reports.append(report)
        # should we send a report to defined callbackers?
        if self._should_send_reports:
            # spin off reporting (no need to wait on it)
            whole_report = DataUpdateReport(update_id=update.id, reports=reports)
            extra_callbacks = self._callbacks_register.normalize_callbacks(update.callback.callbacks)
            asyncio.create_task(self._callbacks_reporter.report_update_results(whole_report, extra_callbacks))



from enum import Enum

from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig
from opal_common.confi import Confi, confi
from opal_common.config import opal_common_config
from opal_common.schemas.data import UpdateCallback
from opal_client.opa.options import OpaServerOptions
from opal_client.policy_store.schemas import PolicyStoreTypes


# Opal Client general configuration -------------------------------------------
class OpaLogFormat(str, Enum):
    NONE = "none"  # no opa logs are piped
    MINIMAL = "minimal"  # only the event name is logged
    HTTP = "http"  # tries to extract http method, path and response status code
    FULL = "full"  # logs the entire data dict returned


class OpalClientConfig(Confi):
    # opa client (policy store) configuration
    POLICY_STORE_TYPE = confi.enum("POLICY_STORE_TYPE", PolicyStoreTypes, PolicyStoreTypes.OPA)
    POLICY_STORE_URL = confi.str("POLICY_STORE_URL", f"http://localhost:8181")
    POLICY_STORE_AUTH_TOKEN = confi.str("POLICY_STORE_AUTH_TOKEN", None, description="the authentication (bearer) token OPAL client will use to authenticate against the policy store (i.e: OPA agent)")
    # create an instance of a policy store upon load

    def load_policy_store():
        from opal_client.policy_store import PolicyStoreClientFactory
        return PolicyStoreClientFactory.create()
    # opa runner configuration (OPA can optionally be run by OPAL) ----------------

    # whether or not OPAL should run OPA by itself in the same container
    INLINE_OPA_ENABLED = confi.bool("INLINE_OPA_ENABLED", True)

    # if inline OPA is indeed enabled, user can pass cli options
    # (configuration) that affects how OPA will run
    INLINE_OPA_CONFIG = confi.model(
        "INLINE_OPA_CONFIG",
        OpaServerOptions,
        {},  # defaults are being set according to OpaServerOptions pydantic definitions (see class)
        description="cli options used when running `opa run --server` inline"
    )

    INLINE_OPA_LOG_FORMAT: OpaLogFormat = confi.enum("INLINE_OPA_LOG_FORMAT", OpaLogFormat, OpaLogFormat.NONE)

    # configuration for fastapi routes
    ALLOWED_ORIGINS = ["*"]

    # general configuration for pub/sub clients
    KEEP_ALIVE_INTERVAL = confi.int("KEEP_ALIVE_INTERVAL", 0)

    # Opal Server general configuration -------------------------------------------

    # opal server url
    SERVER_URL = confi.str("SERVER_URL", "http://localhost:7002", flags=["-s"])
    # opal server pubsub url
    OPAL_WS_ROUTE = "/ws"
    SERVER_WS_URL = confi.str("SERVER_WS_URL", confi.delay(
        lambda SERVER_URL="": SERVER_URL.replace("https", "wss").replace("http", "ws")))
    SERVER_PUBSUB_URL = confi.str("SERVER_PUBSUB_URL", confi.delay("{SERVER_WS_URL}" + f"{OPAL_WS_ROUTE}"))

    # opal server auth token
    CLIENT_TOKEN = confi.str("CLIENT_TOKEN", "THIS_IS_A_DEV_SECRET",
                             description="opal server auth token", flags=["-t"])

    # client-api server
    CLIENT_API_SERVER_WORKER_COUNT = confi.int("CLIENT_API_SERVER_WORKER_COUNT", 1,
                                               description="(if run via CLI) Worker count for the opal-client's internal server")

    CLIENT_API_SERVER_HOST = confi.str("CLIENT_API_SERVER_HOST", "127.0.0.1",
                                       description="(if run via CLI)  Address for the opal-client's internal server to bind")

    CLIENT_API_SERVER_PORT = confi.int("CLIENT_API_SERVER_PORT", 7000,
                                       description="(if run via CLI)  Port for the opal-client's internal server to bind")

    # Policy updater configuration ------------------------------------------------

    # directories in policy repo we should subscribe to for policy code (rego) modules
    POLICY_SUBSCRIPTION_DIRS = confi.list("POLICY_SUBSCRIPTION_DIRS", ["."], delimiter=":",
                                          description="directories in policy repo we should subscribe to")

    # Data updater configuration --------------------------------------------------
    DATA_UPDATER_ENABLED = confi.bool(
        "DATA_UPDATER_ENABLED",
        True,
        description="If set to False, opal client will not listen to dynamic data updates. Dynamic data fetching will be completely disabled.")

    DATA_TOPICS = confi.list("DATA_TOPICS", ["policy_data"],
                             description="Data topics to subscribe to")

    DEFAULT_DATA_SOURCES_CONFIG_URL = confi.str("DEFAULT_DATA_SOURCES_CONFIG_URL", confi.delay("{SERVER_URL}/data/config"),
                                                description="Default URL to fetch data configuration from")

    DEFAULT_DATA_URL = confi.str("DEFAULT_DATA_URL", "http://localhost:8000/policy-config",
                                 description="Default URL to fetch data from")

    SHOULD_REPORT_ON_DATA_UPDATES = confi.bool("SHOULD_REPORT_ON_DATA_UPDATES", False,
                                              description="Should the client report on updates to callbacks defined in DEFAULT_UPDATE_CALLBACKS or within the given updates")
    DEFAULT_UPDATE_CALLBACK_CONFIG = confi.model(
        "DEFAULT_UPDATE_CALLBACK_CONFIG", HttpFetcherConfig, {"method": "post", "headers": {"content-type": "application/json"}, "process_data": False})

    DEFAULT_UPDATE_CALLBACKS = confi.model("DEFAULT_UPDATE_CALLBACKS", UpdateCallback, confi.delay(lambda SERVER_URL="": {
        "callbacks": [f"{SERVER_URL}/data/callback_report"]
    }),
    description="Where/How the client should report on the completion of data updates")

    # OPA transaction log / healthcheck policy ------------------------------------
    OPA_HEALTH_CHECK_POLICY_ENABLED = confi.bool("OPA_HEALTH_CHECK_POLICY_ENABLED", False,
        description="Should we load a special healthcheck policy into OPA that checks " + \
            "that opa was synced correctly and is ready to answer to authorization queries")

    OPA_HEALTH_CHECK_TRANSACTION_LOG_PATH = confi.str(
        "OPA_HEALTH_CHECK_TRANSACTION_LOG_PATH",
        "system/opal/transactions",
        description="Path to OPA document that stores the OPA write transactions")

    OPAL_CLIENT_STAT_ID = confi.str(
        "OPAL_CLIENT_STAT_ID",
        None,
        description="Unique client statistics identifier")

    OPA_HEALTH_CHECK_POLICY_PATH = 'opa/healthcheck/opal.rego'

    def on_load(self):
        # LOGGER
        if self.INLINE_OPA_LOG_FORMAT == OpaLogFormat.NONE:
            opal_common_config.LOG_MODULE_EXCLUDE_LIST.append("opal_client.opa.logger")
            # re-assign to apply to internal confi-entries as well
            opal_common_config.LOG_MODULE_EXCLUDE_LIST = opal_common_config.LOG_MODULE_EXCLUDE_LIST


opal_client_config = OpalClientConfig(prefix="OPAL_")



import asyncio
from typing import Dict, Any, List, Tuple
from opal_common.utils import tuple_to_dict

from opal_client.config import opal_client_config
from opal_common.config import opal_common_config
from opal_common.utils import get_authorization_header
from opal_common.fetcher import FetchingEngine
from opal_common.fetcher.events import FetcherConfig
from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig
from opal_common.logger import logger


class DataFetcher:
    """
    fetches policy data from backend
    """

    def __init__(self,
                 default_data_url: str = None,
                 token: str = None):
        """

        Args:
            default_data_url (str, optional): The URL used to fetch data if no specific url is given in a fetch request. Defaults to DEFAULT_DATA_URL.
            token (str, optional): default auth token. Defaults to CLIENT_TOKEN.
        """
        # defaults
        default_data_url: str = default_data_url or opal_client_config.DEFAULT_DATA_URL
        token: str = token or opal_client_config.CLIENT_TOKEN
        # The underlying fetching engine
        self._engine = FetchingEngine(
            worker_count=opal_common_config.FETCHING_WORKER_COUNT,
            callback_timeout=opal_common_config.FETCHING_CALLBACK_TIMEOUT,
            enqueue_timeout=opal_common_config.FETCHING_ENQUEUE_TIMEOUT
        )
        self._data_url = default_data_url
        self._token = token
        self._auth_headers = tuple_to_dict(get_authorization_header(token))
        self._default_fetcher_config = HttpFetcherConfig(headers=self._auth_headers, is_json=True)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        Context handler to terminate internal tasks
        """
        await self.stop()

    async def start(self):
        self._engine.start_workers()

    async def stop(self):
        """
        Release internal tasks and resources
        """
        await self._engine.terminate_workers()

    async def handle_url(self, url, config):
        """
        Helper function wrapping self._engine.handle_url
        """
        logger.info("Fetching data from url: {url}", url=url)
        try:
            # ask the engine to get our data
            response = await self._engine.handle_url(url, config=config)
            return response
        except asyncio.TimeoutError as e:
            logger.exception("Timeout while fetching url: {url}", url=url)
            raise

    async def handle_urls(self, urls: List[Tuple[str, FetcherConfig]] = None) -> List[Tuple[str, FetcherConfig, Any]]:
        """
        Fetch data for each given url with the (optional) fetching configuration; return the resulting data mapped to each URL

        Args:
            urls (List[Tuple[str, FetcherConfig]], optional): Urls (and fetching configuration) to fetch from.
            Defaults to None - init data_url with HttpFetcherConfig (loaded with the provided auth token).

        Returns:
            List[Tuple[str,FetcherConfig, Any]]: urls mapped to their resulting fetched data
        """

        # tasks
        tasks = []
        # if no url provided - default to the builtin route
        if urls is None:
            urls = [(self._data_url, self._default_fetcher_config)]
        # create a task for each url
        for url, config in urls:
            tasks.append(self.handle_url(url, config))
        # wait for all data fetches to complete
        results = await asyncio.gather(*tasks, return_exceptions=True)

        # Map results with their matching urls and config
        results_with_url_and_config = [(url, config, result) for (url, config), result in zip(urls, results)]

        # return results
        return results_with_url_and_config



import json
from datetime import datetime
from typing import Any, Dict, Optional, List, Union
import uuid
from pydantic import BaseModel
from opal_common.schemas.policy import PolicyBundle
from opal_common.schemas.store import RemoteStatus, StoreTransaction
from inspect import signature
from functools import partial

from opal_client.logger import logger
from opal_client.config import opal_client_config

JsonableValue = Union[Dict[str, Any], List[Any]]

class AbstractPolicyStore:
    """
    holds only the interface of a policy store
    """
    async def set_policy(self, policy_id: str, policy_code: str, transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def get_policy(self, policy_id: str) -> Optional[str]:
        raise NotImplementedError()

    async def delete_policy(self, policy_id: str, transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def get_policy_module_ids(self) -> List[str]:
        raise NotImplementedError()

    async def set_policies(self, bundle: PolicyBundle, transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def get_policy_version(self) -> Optional[str]:
        raise NotImplementedError()

    async def set_policy_data(self, policy_data: JsonableValue, path: str = "", transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def delete_policy_data(self, path: str = "", transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def patch_data(self, path: str, patch_document: Dict[str, Any], transaction_id:Optional[str]=None):
        raise NotImplementedError()

    async def get_data(self, path: str) -> Dict:
        raise NotImplementedError()

    async def get_data_with_input(self, path: str, input: BaseModel) -> Dict:
        raise NotImplementedError()

    async def init_healthcheck_policy(self, policy_id: str, policy_code: str, data_updater_enabled: bool = True):
        raise NotImplementedError()

    async def persist_transaction(self, transaction: StoreTransaction):
        raise NotImplementedError()


class PolicyStoreTransactionContextManager(AbstractPolicyStore):

    def __init__(self, policy_store:"BasePolicyStoreClient", transaction_id=None, transaction_type=None, creation_time=None) -> None:
        self._store = policy_store
        # make sure we have  a transaction id
        self._transaction_id = transaction_id or uuid.uuid4().hex
        self._actions = []
        self._remotes_status: List[RemoteStatus] = []
        self._transaction_type = transaction_type
        self._creation_time = datetime.utcnow().isoformat()

    def __getattribute__(self, name: str) -> Any:
        # internal members are prefixed with '-'
        if name.startswith("_"):
            # return internal members as is
            return super().__getattribute__(name)
        else:
            #proxy to wrapped store
            store_attr = getattr(self._store, name)
            # methods that have a transcation id will get it automatically through this proxy
            if callable(store_attr) and (
                "transaction_id" in signature(store_attr).parameters
                or hasattr(store_attr, "affects_transaction")
            ):
                # record the call as an action in the transaction
                self._actions.append(name)
                return partial(store_attr, transaction_id=self._transaction_id)
            # return properties / and regular methods as is
            else:
                return store_attr

    def _update_remote_status(self, url: str, status: bool, error: str):
        self._remotes_status.append({'remote_url': url, 'succeed': status, 'error': error})

    async def __aenter__(self):
        await self._store.start_transaction(transaction_id=self._transaction_id)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._store.end_transcation(exc_type, exc, tb, transaction_id=self._transaction_id, actions=self._actions,
            transaction_type=self._transaction_type, remotes_status=self._remotes_status, creation_time=self._creation_time)


class BasePolicyStoreClient(AbstractPolicyStore):
    """
    An interface for policy and policy-data store
    """

    def transaction_context(self, transaction_id: str, transaction_type: str) -> PolicyStoreTransactionContextManager:
        """
        Args:
            transaction_id : the id of the transaction

        Returns:
            PolicyStoreTranscationContextManager: a context manager for a transaction to be used in a async with statement
        """
        return PolicyStoreTransactionContextManager(self, transaction_id=transaction_id, transaction_type=transaction_type)

    async def start_transaction(self, transaction_id:str=None):
        """
        PolicyStoreTranscationContextManager calls here on __aenter__
        Start a series of operations with the policy store
        """
        pass

    async def end_transcation(self, exc_type=None, exc=None, tb=None, transaction_id: str = None, actions: List[str] = None,
                            transaction_type: str = None, remotes_status: Optional[List[RemoteStatus]] = None, creation_time = None):
        """
        PolicyStoreTranscationContextManager calls here on __aexit__
        Complete a series of operations with the policy store

        Args:
            exc_type: The exception type (if raised). Defaults to None.
            exc: The exception type (if raised). Defaults to None.
            tb: The traceback (if raised). Defaults to None.
            transaction_id (str, optional): The transaction id. Defaults to None.
            actions (List[str], optional): All the methods called in the transaction. Defaults to None.
        """
        exception_fetching_transaction = []
        if remotes_status and len(remotes_status):
            exception_fetching_transaction = [remote for remote in remotes_status if not remote['succeed']]
        elif transaction_id is None or not actions:
            return # skip, nothing to do if we have no data to log

        end_time = datetime.utcnow().isoformat()
        if exc is not None or len(exception_fetching_transaction):
            try:
                if exc is not None:
                    error_message = repr(exc)
                elif len(exception_fetching_transaction) > 0:
                    network_errors = [remote.get("error", None) for remote in exception_fetching_transaction]
                    network_errors = [str(err) for err in network_errors if err is not None]
                    error_message = ";".join(network_errors) if network_errors else None
                else:
                    error_message = None
            except: # maybe repr throws here
                error_message = None

            transaction = StoreTransaction(
                id=transaction_id,
                actions=actions,
                success=False,
                error=error_message or "",
                creation_time=creation_time,
                end_time=end_time,
                transaction_type=transaction_type,
                remotes_status=remotes_status
            )

            logger.error("OPA transaction failed, transaction id={id}, actions={actions}, error={err}",
                id=transaction_id,
                actions=repr(actions),
                err=error_message
            )
        else:
            transaction = StoreTransaction(id=transaction_id, actions=actions, success=True, creation_time=creation_time, end_time=end_time,
                                            transaction_type=transaction_type, remotes_status=remotes_status)

        if not opal_client_config.OPA_HEALTH_CHECK_POLICY_ENABLED:
            return # skip persisting the transaction, healthcheck policy is disabled

        try:
            await self.persist_transaction(transaction)
        except Exception as e:
            # The writes to transaction log in OPA cache are not done a protected
            # transaction context. If they fail, we do nothing special.
            transaction_data = json.dumps(transaction, indent=4, sort_keys=True, default=str)
            logger.error("Cannot write to OPAL transaction log, transaction id={id}, error={err} with data={data}",
                id=transaction.id,
                err=repr(e),
                data=transaction_data
            )


import asyncio
import os
import sys
from multiprocessing import Event, Process

import pytest
import uvicorn
from aiohttp import ClientSession
from fastapi_websocket_pubsub import PubSubClient
from fastapi_websocket_rpc.logger import LoggingModes, logging_config


# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))
sys.path.append(root_dir)


from opal_client import OpalClient
from opal_client.data.rpc import TenantAwareRpcEventClientMethods
from opal_client.data.updater import DataSourceEntry, DataUpdate, DataUpdater
from opal_client.policy_store import PolicyStoreClientFactory, PolicyStoreTypes
from opal_client.policy_store.mock_policy_store_client import \
    MockPolicyStoreClient
from opal_common.schemas.data import DataSourceConfig, ServerDataSourceConfig
from opal_common.utils import get_authorization_header
from opal_server.config import opal_server_config
from opal_server.server import OpalServer




# Server settings
PORT = int(os.environ.get("PORT") or "9123")
UPDATES_URL = f"ws://localhost:{PORT}/ws"
DATA_ROUTE = "/fetchable_data"
DATA_URL = f"http://localhost:{PORT}{DATA_ROUTE}"
DATA_CONFIG_URL = f"http://localhost:{PORT}{opal_server_config.DATA_CONFIG_ROUTE}"
DATA_TOPICS = ["policy_data"]
TEST_DATA = {
    "hello": "world"
}
DATA_SOURCES_CONFIG = ServerDataSourceConfig(
    config=DataSourceConfig(
        entries=[
            {"url": DATA_URL, "topics": DATA_TOPICS}
        ]
    )
)

# Client settings
CLIENT_PORT = int(os.environ.get("CLIENT_PORT") or "9321")
CLIENT_STORE_ROUTE = "/check_store"
CLIENT_STORE_URL = f"http://localhost:{CLIENT_PORT}{CLIENT_STORE_ROUTE}"


def setup_server(event):
    # Server without git watcher and with a test specifc url for data, and without broadcasting
    server = OpalServer(init_policy_watcher=False, init_publisher=False, data_sources_config=DATA_SOURCES_CONFIG, broadcaster_uri=None, enable_jwks_endpoint=False)
    server_app = server.app

    # add a url to fetch data from
    @server_app.get(DATA_ROUTE)
    def fetchable_data():
        return TEST_DATA

    @server_app.on_event("startup")
    async def startup_event():
        await asyncio.sleep(0.2)
        # signal the server is ready
        event.set()

    uvicorn.run(server_app, port=PORT)


def setup_client(event):

        # config to use mock OPA
        policy_store = PolicyStoreClientFactory.create(store_type=PolicyStoreTypes.MOCK)
        data_updater = DataUpdater(pubsub_url=UPDATES_URL,
                                data_sources_config_url=DATA_CONFIG_URL,
                                policy_store=policy_store,
                                fetch_on_connect=True,
                                data_topics=DATA_TOPICS)

        client = OpalClient(
            policy_store_type=PolicyStoreTypes.MOCK,
            policy_store=policy_store,
            data_updater=data_updater,
            policy_updater=False
        )

        # add a url to fetch data from
        @client.app.get(CLIENT_STORE_ROUTE)
        async def get_store_data():
            # await asyncio.wait_for(client.policy_store.wait_for_data(),5)
            return await client.policy_store.get_data()

        @client.app.on_event("startup")
        async def startup_event():
            store:MockPolicyStoreClient = client.policy_store
            await store.wait_for_data()
            # signal the client is ready
            event.set()

        uvicorn.run(client.app, port=CLIENT_PORT)



@pytest.fixture(scope="module")
def server():
    event = Event()
    # Run the server as a separate process
    proc = Process(target=setup_server, args=(event,), daemon=True)
    proc.start()
    yield event
    proc.kill()  # Cleanup after test


@pytest.fixture(scope="module")
def client():
    event = Event()
    # Run the server as a separate process
    proc = Process(target=setup_client, args=(event,), daemon=True)
    proc.start()
    yield event
    proc.kill()  # Cleanup after test


@pytest.mark.asyncio
async def test_client_connect_to_server_data_updates(client, server):
    """
    Disable auto-update on connect (fetch_on_connect=False)
    Connect to OPAL-server trigger a Data-update and check our policy store gets the update
    """
    server.wait(5)
    client.wait(5)

    async with ClientSession() as session:
        res = await session.get(CLIENT_STORE_URL)
        data = await res.json()
        assert len(data) > 0



import aiohttp

from typing import List, Optional
from pydantic import ValidationError
from fastapi import status, HTTPException
from tenacity import retry, wait, stop

from opal_common.utils import throw_if_bad_status_code, tuple_to_dict
from opal_common.utils import get_authorization_header
from opal_common.security.sslcontext import get_custom_ssl_context
from opal_common.schemas.policy import PolicyBundle
from opal_client.logger import logger
from opal_client.config import opal_client_config


def force_valid_bundle(bundle) -> PolicyBundle:
    try:
        return PolicyBundle(**bundle)
    except ValidationError as e:
        logger.warning("server returned invalid bundle: {err}", bundle=bundle, err=repr(e))
        raise


class PolicyFetcher:
    """
    fetches policy from backend
    """
    DEFAULT_RETRY_CONFIG = {
        'wait': wait.wait_random_exponential(max=10),
        'stop': stop.stop_after_attempt(5),
        'reraise': True,
    }

    def __init__(self, backend_url=None, token=None, retry_config=None):
        """
        Args:
            backend_url (str): Defaults to opal_client_config.SERVER_URL.
            token ([type], optional): [description]. Defaults to opal_client_config.CLIENT_TOKEN.
        """
        self._token = token or opal_client_config.CLIENT_TOKEN
        self._backend_url = backend_url or opal_client_config.SERVER_URL
        self._auth_headers = tuple_to_dict(get_authorization_header(self._token))
        self._retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG
        self._policy_endpoint_url = f"{self._backend_url}/policy"
        # custom SSL context (for self-signed certificates)
        self._custom_ssl_context = get_custom_ssl_context()
        self._ssl_context_kwargs = {'ssl': self._custom_ssl_context} if self._custom_ssl_context is not None else {}

    @property
    def policy_endpoint_url(self):
        return self._policy_endpoint_url

    async def fetch_policy_bundle(
        self,
        directories: List[str] = ['.'],
        base_hash: Optional[str] = None
    ) -> Optional[PolicyBundle]:
        attempter = retry(**self._retry_config)(self._fetch_policy_bundle)
        try:
            return await attempter(directories=directories, base_hash=base_hash)
        except Exception as err:
            logger.warning("Failed all attempts to fetch bundle, got error: {err}", err=repr(err))
            raise

    async def _fetch_policy_bundle(
        self,
        directories: List[str] = ['.'],
        base_hash: Optional[str] = None
    ) -> Optional[PolicyBundle]:
        """
        Fetches the bundle. May throw, in which case we retry again.
        """
        params = {"path": directories}
        if base_hash is not None:
            params["base_hash"] = base_hash
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    self._policy_endpoint_url,
                    headers={'content-type': 'text/plain', **self._auth_headers},
                    params=params,
                    **self._ssl_context_kwargs
                ) as response:
                    if response.status == status.HTTP_404_NOT_FOUND:
                        logger.warning("requested paths not found: {paths}", paths=directories)
                        raise HTTPException(
                            status_code=status.HTTP_404_NOT_FOUND,
                            detail=f"requested path {self._policy_endpoint_url} was not found in the policy repo!"
                        )


                    # may throw ValueError
                    await throw_if_bad_status_code(response, expected=[status.HTTP_200_OK], logger=logger)

                    # may throw Validation Error
                    bundle = await response.json()
                    return force_valid_bundle(bundle)
            except aiohttp.ClientError as e:
                logger.warning("server connection error: {err}", err=repr(e))
                raise




import asyncio
import logging
import multiprocessing
import os
import sys
from multiprocessing import Event, Process
import pytest
import uvicorn
from fastapi_websocket_pubsub import PubSubClient
from aiohttp import ClientSession

# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir))
sys.path.append(root_dir)

from opal_client.config import opal_client_config
from opal_client.data.rpc import TenantAwareRpcEventClientMethods
from opal_client.data.updater import DataSourceEntry, DataUpdate, DataUpdater
from opal_client.policy_store import (PolicyStoreClientFactory,
                                      PolicyStoreTypes)
from opal_common.schemas.data import DataSourceConfig, DataUpdateReport, ServerDataSourceConfig, UpdateCallback
from opal_common.utils import get_authorization_header
from opal_server.config import opal_server_config
from opal_server.server import OpalServer



PORT = int(os.environ.get("PORT") or "9123")
UPDATES_URL = f"ws://localhost:{PORT}/ws"
DATA_ROUTE = "/fetchable_data"
DATA_URL = f"http://localhost:{PORT}{DATA_ROUTE}"
DATA_CONFIG_URL = f"http://localhost:{PORT}{opal_server_config.DATA_CONFIG_ROUTE}"
DATA_TOPICS = ["policy_data"]
TEST_DATA = {
    "hello": "world"
}

DATA_UPDATE_CALLBACK_ROUTE = "/data/callback_report_for_test"
DATA_UPDATE_CALLBACK_URL = f"http://localhost:{PORT}{DATA_UPDATE_CALLBACK_ROUTE}"

CHECK_DATA_UPDATE_CALLBACK_ROUTE = "/callback_count"
CHECK_DATA_UPDATE_CALLBACK_URL = f"http://localhost:{PORT}{CHECK_DATA_UPDATE_CALLBACK_ROUTE}"

DATA_SOURCES_CONFIG = ServerDataSourceConfig(
    config=DataSourceConfig(
        entries=[
            {"url": DATA_URL, "topics": DATA_TOPICS}
        ]
    )
)


def setup_server(event):
    # Server without git watcher and with a test specifc url for data, and without broadcasting
    server = OpalServer(init_policy_watcher=False, data_sources_config=DATA_SOURCES_CONFIG, broadcaster_uri=None, enable_jwks_endpoint=False)
    server_app = server.app

    callbacks = []

    # add a url to fetch data from
    @server_app.get(DATA_ROUTE)
    def fetchable_data():
        return TEST_DATA


    # route to report complition to
    @server_app.post(DATA_UPDATE_CALLBACK_ROUTE)
    def callback(report:DataUpdateReport):
        assert report.reports[0].hash == DataUpdater.calc_hash(TEST_DATA)
        callbacks.append(report)
        return "OKAY"


    # route to report complition to
    @server_app.get(CHECK_DATA_UPDATE_CALLBACK_ROUTE)
    def check()->int:
        return len(callbacks)




    @server_app.on_event("startup")
    async def startup_event():
        await asyncio.sleep(0.4)
        # signal the server is ready
        event.set()

    uvicorn.run(server_app, port=PORT)


@pytest.fixture(scope="module")
def server():
    event = Event()
    # Run the server as a separate process
    proc = Process(target=setup_server, args=(event,), daemon=True)
    proc.start()
    yield event
    proc.kill()  # Cleanup after test


def trigger_update():
    async def run():
        # trigger an update
        entries = [DataSourceEntry(url=DATA_URL)]
        callback = UpdateCallback(callbacks=[DATA_UPDATE_CALLBACK_URL])
        update = DataUpdate(reason="Test", entries=entries, callback=callback)
        async with PubSubClient(
            server_uri=UPDATES_URL,
            methods_class=TenantAwareRpcEventClientMethods,
            extra_headers=[get_authorization_header(opal_client_config.CLIENT_TOKEN)]
        ) as client:
            # Channel must be ready before we can publish on it
            await asyncio.wait_for(client.wait_until_ready(), 5)
            logging.info("Publishing data event")
            await client.publish(DATA_TOPICS, data=update)
    asyncio.run(run())


@pytest.mark.asyncio
async def test_data_updater(server):
    """
    Disable auto-update on connect (fetch_on_connect=False)
    Connect to OPAL-server trigger a Data-update and check our policy store gets the update
    """
    # Wait for the server to start
    server.wait(5)
    # config to use mock OPA
    policy_store = PolicyStoreClientFactory.create(store_type=PolicyStoreTypes.MOCK)
    updater = DataUpdater(pubsub_url=UPDATES_URL, policy_store=policy_store,
                          fetch_on_connect=False, data_topics=DATA_TOPICS, should_send_reports=False)
    # start the updater (terminate on exit)
    await updater.start()
    try:
        proc = multiprocessing.Process(target=trigger_update, daemon=True)
        proc.start()
        # wait until new data arrives into the strore via the updater
        await asyncio.wait_for(policy_store.wait_for_data(), 60)
    # cleanup
    finally:
        await updater.stop()
        proc.terminate()

@pytest.mark.asyncio
async def test_data_updater_with_report_callback(server):
    """
    Disable auto-update on connect (fetch_on_connect=False)
    Connect to OPAL-server trigger a Data-update and check our policy store gets the update
    """
    # Wait for the server to start
    server.wait(5)
    # config to use mock OPA
    policy_store = PolicyStoreClientFactory.create(store_type=PolicyStoreTypes.MOCK)
    updater = DataUpdater(pubsub_url=UPDATES_URL, policy_store=policy_store,
                          fetch_on_connect=False, data_topics=DATA_TOPICS, should_send_reports=True)
    # start the updater (terminate on exit)
    await updater.start()

    current_callback_count = 0
    async with ClientSession() as session:
        res = await session.get(CHECK_DATA_UPDATE_CALLBACK_URL)
        current_callback_count = await res.json()

    try:
        proc = multiprocessing.Process(target=trigger_update, daemon=True)
        proc.start()
        # wait until new data arrives into the strore via the updater
        await asyncio.wait_for(policy_store.wait_for_data(), 15)
        # give the callback a chance to arrive
        await asyncio.sleep(1)

        async with ClientSession() as session:
            res = await session.get(CHECK_DATA_UPDATE_CALLBACK_URL)
            count = await res.json()
            # we got one callback in the interim
            assert count == current_callback_count +1

    # cleanup
    finally:
        await updater.stop()
        proc.terminate()


@pytest.mark.asyncio
async def test_client_get_initial_data(server):
    """
    Connect to OPAL-server and make sure data is fetched on-connect
    """
    # Wait for the server to start
    server.wait(5)
    # config to use mock OPA
    policy_store = PolicyStoreClientFactory.create(store_type=PolicyStoreTypes.MOCK)
    updater = DataUpdater(pubsub_url=UPDATES_URL, data_sources_config_url=DATA_CONFIG_URL,
                          policy_store=policy_store, fetch_on_connect=True, data_topics=DATA_TOPICS, should_send_reports=False)
    # start the updater (terminate on exit)
    await updater.start()
    try:
        # wait until new data arrives into the strore via the updater
        await asyncio.wait_for(policy_store.wait_for_data(), 5)
    # cleanup
    finally:
        await updater.stop()


from .mock_policy_store_client import MockPolicyStoreClient
from .opa_client import OpaClient
from .base_policy_store_client import BasePolicyStoreClient
from .policy_store_client_factory import PolicyStoreClientFactory, DEFAULT_POLICY_STORE_GETTER, PolicyStoreTypes


import json
import logging

from typing import Optional
from enum import Enum

from opal_client.logger import logger
from opal_client.config import opal_client_config, OpaLogFormat

def logging_level_from_string(level: str) -> int:
    """
    logger.log() requires an int logging level
    """
    level = level.lower()
    if level == "info":
        return logging.INFO
    elif level == "critical":
        return logging.CRITICAL
    elif level == "fatal":
        return logging.FATAL
    elif level == "error":
        return logging.ERROR
    elif level == "warning" or level == "warn":
        return logging.WARNING
    elif level == "debug":
        return logging.DEBUG
    # default
    return logging.INFO

async def pipe_opa_logs(stream):
    """
    gets a stream of logs from the opa process, and logs it into the main opal log.
    """
    if opal_client_config.INLINE_OPA_LOG_FORMAT == OpaLogFormat.NONE:
        return

    while True:
        line = await stream.readline()
        if not line:
            break
        try:
            log_line = json.loads(line)
            level = logging.getLevelName(logging_level_from_string(log_line.pop("level", "info")))
            msg = log_line.pop("msg", None)

            logged = False
            if opal_client_config.INLINE_OPA_LOG_FORMAT == OpaLogFormat.MINIMAL:
                logged = log_event_name(level, msg)
            elif opal_client_config.INLINE_OPA_LOG_FORMAT == OpaLogFormat.HTTP:
                logged = log_formatted_http_details(level, msg, log_line)

            # always fall back to log the entire line
            if not logged or opal_client_config.INLINE_OPA_LOG_FORMAT == OpaLogFormat.FULL:
                log_entire_dict(level, msg, log_line)
        except json.JSONDecodeError:
            logger.info(line)

def log_event_name(level: str, msg: Optional[str]) -> bool:
    if msg is not None:
        logger.log(level, "{msg: <20}", msg=msg)
        return True
    return False

def log_formatted_http_details(level: str, msg: Optional[str], log_line: dict) -> bool:
    method: Optional[str] = log_line.pop("req_method", None)
    path: Optional[str] = log_line.pop("req_path", None)
    status: Optional[int] = log_line.pop("resp_status", None)

    if msg is None or method is None or path is None:
        return False

    if status is None:
        format = "{msg: <20} <fg #999>{method} {path}</>"
        logger.opt(colors=True).log(level, format, msg=msg, method=method, path=path)
    else:
        format = "{msg: <20} <fg #999>{method} {path} -> {status}</>"
        logger.opt(colors=True).log(level, format, msg=msg, method=method, path=path, status=status)

    return True

def log_entire_dict(level: str, msg: Optional[str], log_line: dict):
    if msg is None:
        format = "<fg #999>{log_line}</>"
    else:
        format = "{msg: <20} <fg #bfbfbf>{log_line}</>"

    try:
        log_line = json.dumps(log_line) # should be ok, originated in json
    except:
        pass # fallback to dict
    logger.opt(colors=True).log(level, format, msg=msg, log_line=log_line)
    return True



from typing import Dict, Optional
from opal_client.policy_store.base_policy_store_client import BasePolicyStoreClient
from opal_client.config import opal_client_config, PolicyStoreTypes


class PolicyStoreClientFactoryException(Exception):
    pass


class InvalidPolicyStoreTypeException(Exception):
    pass


class PolicyStoreClientFactory:

    CACHE: Dict[str,BasePolicyStoreClient] = {}

    @classmethod
    def get(cls,store_type: PolicyStoreTypes = None, url: str = None, save_to_cache=True, token:Optional[str]=None) -> BasePolicyStoreClient:
        """Same as self.create() but with caching

        Args:
            store_type (PolicyStoreTypes, optional): The type of policy-store to use. Defaults to opal_client_config.POLICY_STORE_TYPE.
            url (str, optional): the URL of the policy store. Defaults to opal_client_config.POLICY_STORE_URL.
            save_to_cache (bool, optional): Should the created value be saved to cache (To be obtained via the get method).

        Raises:
            InvalidPolicyStoreTypeException: Raised when the factory doesn't have a store-client macthing the given type

        Returns:
            BasePolicyStoreClient: the policy store client interface
        """
        # get from cache if available - else create anew
        key = cls.get_cache_key(store_type,url)
        value = cls.CACHE.get(key, None)
        if value is None:
            return cls.create(store_type,url, token)
        else:
            return value

    @classmethod
    def create(cls, store_type: PolicyStoreTypes = None, url: str = None, save_to_cache=True, token:Optional[str]=None) -> BasePolicyStoreClient:
        """
        Factory method - create a new policy store by type.

        Args:
            store_type (PolicyStoreTypes, optional): The type of policy-store to use. Defaults to opal_client_config.POLICY_STORE_TYPE.
            url (str, optional): the URL of the policy store. Defaults to opal_client_config.POLICY_STORE_URL.
            save_to_cache (bool, optional): Should the created value be saved to cache (To be obtained via the get method).

        Raises:
            InvalidPolicyStoreTypeException: Raised when the factory doesn't have a store-client macthing the given type

        Returns:
            BasePolicyStoreClient: the policy store client interface
        """
        # load defaults
        store_type = store_type or opal_client_config.POLICY_STORE_TYPE
        url = url or opal_client_config.POLICY_STORE_URL
        store_token = token or opal_client_config.POLICY_STORE_AUTH_TOKEN

        # OPA
        if PolicyStoreTypes.OPA == store_type:
            from opal_client.policy_store.opa_client import OpaClient
            res = OpaClient(url, opa_auth_token=store_token)
        # MOCK
        elif PolicyStoreTypes.MOCK == store_type:
            from opal_client.policy_store.mock_policy_store_client import MockPolicyStoreClient
            res = MockPolicyStoreClient()
        else:
            raise InvalidPolicyStoreTypeException(f"{store_type} is not a valid policy store type")
        # save to cache
        if save_to_cache:
            cls.CACHE[cls.get_cache_key(store_type, url)] = res
        # return the result
        return res

    @staticmethod
    def get_cache_key(store_type, url):
        return f"{store_type.value}|{url}"


DEFAULT_POLICY_STORE_GETTER = PolicyStoreClientFactory.get



from fastapi import APIRouter, Depends

from opal_common.logger import logger
from opal_common.schemas.security import PeerType
from opal_common.authentication.deps import JWTAuthenticator
from opal_common.authentication.verifier import Unauthorized
from opal_common.authentication.types import JWTClaims
from opal_common.authentication.authz import require_peer_type
from opal_client.policy_store.schemas import PolicyStoreDetails
from opal_client.config import opal_client_config

def init_policy_store_router(authenticator: JWTAuthenticator):
    router = APIRouter()

    @router.get("/policy-store/config", response_model=PolicyStoreDetails, response_model_exclude_none=True)
    async def get_policy_store_details(claims: JWTClaims = Depends(authenticator)):
        try:
            require_peer_type(authenticator, claims, PeerType.listener) # may throw Unauthorized
        except Unauthorized as e:
            logger.error(f"Unauthorized to publish update: {repr(e)}")
            raise

        return PolicyStoreDetails(
            url=opal_client_config.POLICY_STORE_URL,
            token=opal_client_config.POLICY_STORE_AUTH_TOKEN or None
        )

    return router



import asyncio
import aiohttp
import json
import functools
import time
from typing import Dict, Any, Optional, List, Set

from tenacity import retry, stop_after_attempt, wait_fixed
from pydantic import BaseModel
from fastapi import Response, status

from opal_client.config import opal_client_config
from opal_client.logger import logger
from opal_client.utils import proxy_response
from opal_common.schemas.policy import DataModule, PolicyBundle
from opal_common.schemas.store import JSONPatchAction, StoreTransaction, TransactionType
from opal_common.opa.parsing import get_rego_package
from opal_common.git.bundle_utils import BundleUtils
from opal_client.policy_store.base_policy_store_client import BasePolicyStoreClient, JsonableValue


JSONPatchDocument = List[JSONPatchAction]

# 2 retries with 2 seconds apart
RETRY_CONFIG = dict(wait=wait_fixed(2), stop=stop_after_attempt(2))


def fail_silently(fallback=None):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except aiohttp.ClientError as e:
                return fallback
        return wrapper
    return decorator

def affects_transaction(func):
    """
    mark a method as write (affecting state of transaction) for transaction log
    """
    setattr(func, 'affects_transaction', True)
    return func

async def proxy_response_unless_invalid(raw_response: aiohttp.ClientResponse, accepted_status_codes: List[int]) -> Response:
    """
    throws value error if the http response recieved has an unexpected status code
    """
    response = await proxy_response(raw_response)
    if response.status_code not in accepted_status_codes:
        try:
            error = await raw_response.json()
        except json.JSONDecodeError:
            error = ""
        raise ValueError("OPA Client: unexpected status code: {}, error: {}".format(response.status_code, error))
    return response


class OpaTransactionLogState:
    """
    holds a mutatable state of the transaction log.
    can persist to OPA as hardcoded policy
    """
    POLICY_ACTIONS = ["set_policies", "set_policy", "delete_policy"]
    DATA_ACTIONS = ["set_policy_data", "delete_policy_data"]

    def __init__(self, policy_store: BasePolicyStoreClient, policy_id: str, policy_template: str, data_updater_enabled: bool = True):
        self._store = policy_store
        self._policy_id = policy_id
        self._policy_template = policy_template
        self._data_updater_disabled = not data_updater_enabled
        self._num_successful_policy_transactions = 0
        self._num_failed_policy_transactions = 0
        self._num_successful_data_transactions = 0
        self._num_failed_data_transactions = 0
        self._last_policy_transaction: Optional[StoreTransaction] = None
        self._last_failed_policy_transaction: Optional[StoreTransaction] = None
        self._last_data_transaction: Optional[StoreTransaction] = None
        self._last_failed_data_transaction: Optional[StoreTransaction] = None

    @property
    def ready(self) -> bool:
        is_ready: bool = (
            self._num_successful_policy_transactions > 0 and
            (self._data_updater_disabled or self._num_successful_data_transactions > 0)
        )
        return json.dumps(is_ready)

    @property
    def healthy(self) -> bool:
        policy_updater_is_healthy: bool = (
            self._last_policy_transaction is not None and
            self._last_policy_transaction.success
        )
        data_updater_is_healthy: bool = (
            self._last_data_transaction is not None and
            self._last_data_transaction.success
        )
        is_healthy: bool = policy_updater_is_healthy and (self._data_updater_disabled or data_updater_is_healthy)
        return json.dumps(is_healthy)

    @property
    def last_policy_transaction(self):
        if self._last_policy_transaction is None:
            return json.dumps({})
        return json.dumps(self._last_policy_transaction.dict())

    @property
    def last_data_transaction(self):
        if self._last_data_transaction is None:
            return json.dumps({})
        return json.dumps(self._last_data_transaction.dict())

    @property
    def last_failed_policy_transaction(self):
        if self._last_failed_policy_transaction is None:
            return json.dumps({})
        return json.dumps(self._last_failed_policy_transaction.dict())

    @property
    def last_failed_data_transaction(self):
        if self._last_failed_data_transaction is None:
            return json.dumps({})
        return json.dumps(self._last_failed_data_transaction.dict())

    @property
    def transaction_policy_statistics(self):
        return json.dumps({"successful": self._num_successful_policy_transactions, "failed": self._num_failed_policy_transactions})

    @property
    def transaction_data_statistics(self):
        return json.dumps({"successful": self._num_successful_data_transactions, "failed": self._num_failed_data_transactions})

    async def persist(self):
        """
        renders the policy template with the current state, and writes it to OPA
        """
        logger.info("persisting health check policy: ready={ready}, healthy={healthy}", ready=self.ready, healthy=self.healthy)
        logger.info("Policy and data statistics: policy: (successful {success_policy}, failed {failed_policy});\tdata: (successful {success_data}, failed {failed_data})",
                success_policy=self._num_successful_policy_transactions, failed_policy=self._num_failed_policy_transactions,
                success_data=self._num_successful_data_transactions, failed_data=self._num_failed_data_transactions)
        policy_code = self._policy_template.format(
            ready=self.ready,
            healthy=self.healthy,
            last_policy_transaction=self.last_policy_transaction,
            last_failed_policy_transaction=self.last_failed_policy_transaction,
            last_data_transaction=self.last_data_transaction,
            last_failed_data_transaction=self.last_failed_data_transaction,
            transaction_data_statistics=self.transaction_data_statistics,
            transaction_policy_statistics=self.transaction_policy_statistics,
        )
        return await self._store.set_policy(policy_id=self._policy_id, policy_code=policy_code)

    def _is_policy_transaction(self, transaction: StoreTransaction):
        return transaction.transaction_type == TransactionType.policy

    def _is_data_transaction(self, transaction: StoreTransaction):
        return transaction.transaction_type == TransactionType.data

    def process_transaction(self, transaction: StoreTransaction):
        """
        mutates the state into a new state that can be then persisted as hardcoded policy
        """
        logger.info("processing store transaction: {transaction}", transaction=transaction.dict())
        if self._is_policy_transaction(transaction):

            if transaction.success:
                self._last_policy_transaction = transaction
                self._num_successful_policy_transactions += 1
            else:
                self._last_failed_policy_transaction = transaction
                self._num_failed_policy_transactions += 1

        elif self._is_data_transaction(transaction):

            if transaction.success:
                self._last_data_transaction = transaction
                self._num_successful_data_transactions += 1
            else:
                self._last_failed_data_transaction = transaction
                self._num_failed_data_transactions += 1

class OpaClient(BasePolicyStoreClient):
    """
    communicates with OPA via its REST API.
    """
    POLICY_NAME = "rbac"

    def __init__(self, opa_server_url=None, opa_auth_token:Optional[str]=None):
        base_url = opa_server_url or opal_client_config.POLICY_STORE_URL
        self._opa_url = f"{base_url}/v1"
        self._cached_policies: Dict[str, str] = {}
        self._policy_version: Optional[str] = None
        self._lock = asyncio.Lock()
        self._token = opa_auth_token
        self._headers = {}
        if opa_auth_token is not None:
            self._headers.update({"Authorization": f"Bearer {opa_auth_token}"})

        # as long as this is null, transaction log is disabled
        self._transaction_state: Optional[OpaTransactionLogState] = None

    async def get_policy_version(self) -> Optional[str]:
        return self._policy_version

    @affects_transaction
    @retry(**RETRY_CONFIG)
    async def set_policy(self, policy_id: str, policy_code: str, transaction_id:Optional[str]=None):
        self._cached_policies[policy_id] = policy_code
        async with aiohttp.ClientSession() as session:
            try:
                async with session.put(
                    f"{self._opa_url}/policies/{policy_id}",
                    data=policy_code,
                    headers={'content-type': 'text/plain', **self._headers}
                ) as opa_response:
                    return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[status.HTTP_200_OK])
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @fail_silently()
    @retry(**RETRY_CONFIG)
    async def get_policy(self, policy_id: str) -> Optional[str]:
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    f"{self._opa_url}/policies/{policy_id}",
                    headers=self._headers
                ) as opa_response:
                    result = await opa_response.json()
                    return result.get("result", {}).get("raw", None)
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @affects_transaction
    @retry(**RETRY_CONFIG)
    async def delete_policy(self, policy_id: str, transaction_id:Optional[str]=None):
        async with aiohttp.ClientSession() as session:
            try:
                async with session.delete(
                    f"{self._opa_url}/policies/{policy_id}",
                    headers=self._headers
                ) as opa_response:
                    return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                        status.HTTP_200_OK,
                        status.HTTP_404_NOT_FOUND
                    ])
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @fail_silently()
    @retry(**RETRY_CONFIG)
    async def get_policy_module_ids(self) -> List[str]:
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    f"{self._opa_url}/policies",
                    headers=self._headers
                ) as opa_response:
                    result = await opa_response.json()
                    return OpaClient._extract_module_ids_from_policies_json(result)
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @staticmethod
    def _extract_module_ids_from_policies_json(result: Dict[str, Any]) -> List[str]:
        """
        return all module ids in OPA cache who are not:
        - skipped module ids (i.e: our health check policy)
        - all modules with package name starting with "system" (special OPA policies)
        """
        modules: List[Dict[str, Any]] = result.get("result", [])
        builtin_modules = [opal_client_config.OPA_HEALTH_CHECK_POLICY_PATH]

        module_ids = []
        for module in modules:
            module_id = module.get("id", None)
            package_name = get_rego_package(module.get("raw", ""))

            if module_id is None:
                continue

            if package_name is not None and package_name.startswith("system."):
                continue

            if module_id in builtin_modules:
                continue

            module_ids.append(module_id)

        return module_ids

    @affects_transaction
    async def set_policies(self, bundle: PolicyBundle, transaction_id:Optional[str]=None):
        if bundle.old_hash is None:
            return await self._set_policies_from_complete_bundle(bundle)
        else:
            return await self._set_policies_from_delta_bundle(bundle)

    async def _set_policies_from_complete_bundle(self, bundle: PolicyBundle):
        module_ids_in_store: Set[str] = set(await self.get_policy_module_ids())
        module_ids_in_bundle: Set[str] = {module.path for module in bundle.policy_modules}
        module_ids_to_delete: Set[str] = module_ids_in_store.difference(module_ids_in_bundle)

        async with self._lock:
            # save bundled policy *static* data into store
            for module in BundleUtils.sorted_data_modules_to_load(bundle):
                await self._set_policy_data_from_bundle_data_module(module, hash=bundle.hash)

            # save bundled policies into store
            for module in BundleUtils.sorted_policy_modules_to_load(bundle):
                await self.set_policy(policy_id=module.path, policy_code=module.rego)

            # remove policies from the store that are not in the bundle
            # (because this bundle is "complete", i.e: contains all policy modules for a given hash)
            for module_id in module_ids_to_delete:
                await self.delete_policy(policy_id=module_id)

            # save policy version (hash) into store
            self._policy_version = bundle.hash

    async def _set_policies_from_delta_bundle(self, bundle: PolicyBundle):
        async with self._lock:
            # save bundled policy *static* data into store
            for module in BundleUtils.sorted_data_modules_to_load(bundle):
                await self._set_policy_data_from_bundle_data_module(module, hash=bundle.hash)

            # save bundled policies into store
            for module in BundleUtils.sorted_policy_modules_to_load(bundle):
                await self.set_policy(policy_id=module.path, policy_code=module.rego)

            # remove deleted policies (or static policy data) from store
            if bundle.deleted_files is not None:
                for module_id in BundleUtils.sorted_policy_modules_to_delete(bundle):
                    await self.delete_policy(policy_id=module_id)

                for module_id in BundleUtils.sorted_data_modules_to_delete(bundle):
                    await self.delete_policy_data(path=self._safe_data_module_path(str(module_id)))

            # save policy version (hash) into store
            self._policy_version = bundle.hash

    @classmethod
    def _safe_data_module_path(cls, path: str):
        if not path or path == ".":
            return ""

        if not path.startswith("/"):
            return f"/{path}"

        return path

    async def _set_policy_data_from_bundle_data_module(self, module: DataModule, hash: Optional[str] = None):
        module_path = self._safe_data_module_path(module.path)
        try:
            module_data = json.loads(module.data)
            return await self.set_policy_data(
                policy_data=module_data,
                path=module_path,
            )
        except aiohttp.ClientError as e:
            logger.warning("Opa connection error: {err}", err=repr(e))
            raise
        except json.JSONDecodeError as e:
            logger.warning(
                "bundle contains non-json data module: {module_path}",
                err=repr(e),
                module_path=module_path,
                bundle_hash=hash
            )

    @affects_transaction
    @retry(**RETRY_CONFIG)
    async def set_policy_data(self, policy_data: JsonableValue, path: str = "", transaction_id:Optional[str]=None):
        path = self._safe_data_module_path(path)

        # in OPA, the root document must be an object, so we must wrap list values
        if not path and isinstance(policy_data, list):
            logger.warning("OPAL client was instructed to put a list on OPA's root document. In OPA the root document must be an object so the original value was wrapped.")
            policy_data = {"items": policy_data}

        async with aiohttp.ClientSession() as session:
            try:
                async with session.put(
                    f"{self._opa_url}/data{path}",
                    data=json.dumps(policy_data, default=str),
                    headers=self._headers
                ) as opa_response:
                    return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                        status.HTTP_204_NO_CONTENT,
                        status.HTTP_304_NOT_MODIFIED
                    ])
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @affects_transaction
    @retry(**RETRY_CONFIG)
    async def delete_policy_data(self, path: str = "", transaction_id:Optional[str]=None):
        path = self._safe_data_module_path(path)
        if not path:
            return await self.set_policy_data({})

        async with aiohttp.ClientSession() as session:
            try:
                async with session.delete(
                    f"{self._opa_url}/data{path}",
                    headers=self._headers
                ) as opa_response:
                    return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[
                        status.HTTP_204_NO_CONTENT,
                        status.HTTP_404_NOT_FOUND
                    ])
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @affects_transaction
    async def patch_data(self, path: str, patch_document: JSONPatchDocument, transaction_id:Optional[str]=None):
        path = self._safe_data_module_path(path)
        # a patch document is a list of actions
        # we render each action with pydantic, and then dump the doc into json
        json_document = json.dumps([action.dict() for action in patch_document])

        async with aiohttp.ClientSession() as session:
            try:
                async with session.patch(
                    f"{self._opa_url}/data{path}",
                    data=json_document,
                    headers=self._headers
                ) as opa_response:
                    return await proxy_response_unless_invalid(opa_response, accepted_status_codes=[status.HTTP_204_NO_CONTENT])
            except aiohttp.ClientError as e:
                logger.warning("Opa connection error: {err}", err=repr(e))
                raise

    @fail_silently()
    @retry(**RETRY_CONFIG)
    async def get_data(self, path: str) -> Dict:
        """
        wraps opa's "GET /data" api that extracts base data documents from opa cache.
        NOTE: opa always returns 200 and empty dict (for valid input) even if the data does not exist.

        returns a dict (parsed json).
        """
        # function accepts paths that start with / and also path that do not start with /
        if path.startswith("/"):
            path = path[1:]
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    f"{self._opa_url}/data/{path}",
                    headers=self._headers
                ) as opa_response:
                    return await opa_response.json()
        except aiohttp.ClientError as e:
            logger.warning("Opa connection error: {err}", err=repr(e))
            raise

    @retry(**RETRY_CONFIG)
    async def get_data_with_input(self, path: str, input: BaseModel) -> Dict:
        """
        evaluates a data document against an input.
        that is how OPA "runs queries".

        see explanation how opa evaluate documents:
        https://www.openpolicyagent.org/docs/latest/philosophy/#the-opa-document-model

        see api reference:
        https://www.openpolicyagent.org/docs/latest/rest-api/#get-a-document-with-input
        """
        # opa data api format needs the input to sit under "input"
        opa_input = {
            "input": input.dict()
        }
        if path.startswith("/"):
            path = path[1:]
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{self._opa_url}/data/{path}",
                    data=json.dumps(opa_input),
                    headers=self._headers
                ) as opa_response:
                    return await proxy_response(opa_response)
        except aiohttp.ClientError as e:
            logger.warning("Opa connection error: {err}", err=repr(e))
            raise

    @retry(**RETRY_CONFIG)
    async def init_healthcheck_policy(self, policy_id: str, policy_code: str, data_updater_enabled: bool = True):
        self._transaction_state = OpaTransactionLogState(
            policy_store=self,
            policy_id=policy_id,
            policy_template=policy_code,
            data_updater_enabled=data_updater_enabled
        )
        return await self._transaction_state.persist()

    @retry(**RETRY_CONFIG)
    async def persist_transaction(self, transaction: StoreTransaction):
        if self._transaction_state is None:
            return
        self._transaction_state.process_transaction(transaction)
        return await self._transaction_state.persist()



import asyncio
import psutil
import time

from typing import Callable, Coroutine, Optional, List

from tenacity import retry, wait_random_exponential

from opal_client.logger import logger
from opal_client.opa.options import OpaServerOptions
from opal_client.opa.logger import pipe_opa_logs


AsyncCallback = Callable[[], Coroutine]

async def wait_until_process_is_up(process_pid: int, callback: Optional[AsyncCallback], wait_interval: float=0.1, timeout: Optional[float] = None):
    """
    waits until the pid of the process exists, then optionally runs a callback.
    optionally receives a timeout to give up.
    """
    start_time = time.time()
    while not psutil.pid_exists(process_pid):
        if timeout is not None and start_time - time() > timeout:
            break
        await asyncio.sleep(wait_interval)
    if callback is not None:
        await callback()

class OpaRunner:
    """
    Runs OPA in a supervised subprocess.

    - if the OPA process fails, OPA runner will restart the process.
    - OPA runner can register callbacks on the lifecycle of OPA,
    making it easy to keep the OPA cache hydrated (up-to-date).
    - OPA runner can pipe the logs of the OPA process into OPAL logger.
    """
    def __init__(self, options: Optional[OpaServerOptions] = None):
        self._options = options or OpaServerOptions()
        self._stopped = False
        self._process = None
        self._should_stop: Optional[asyncio.Event] = None
        self._run_task: Optional[asyncio.Task] = None
        self._on_opa_initial_start_callbacks: List[AsyncCallback] = []
        self._on_opa_restart_callbacks: List[AsyncCallback] = []
        self._process_was_never_up_before = True

    async def __aenter__(self):
        self.start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.stop()

    def start(self):
        """
        starts the opa runner task, and launches the OPA subprocess
        """
        logger.info("Launching opa runner")
        self._run_task = asyncio.create_task(self._run())

    async def stop(self):
        """
        stops the opa runner task (and terminates OPA)
        """
        self._init_events()
        if not self._should_stop.is_set():
            logger.info("Stopping opa runner")
            self._should_stop.set()
            self._terminate_opa()
            await asyncio.sleep(1) # wait for opa process to go down

        if self._run_task is not None:
            await self._run_task
        self._run_task = None

    async def wait_until_done(self):
        """
        waits until the OPA runner task is complete.
        this is great when using opa runner as a context manager.
        """
        if self._run_task is not None:
            await self._run_task

    @property
    def command(self):
        opts = self._options.get_cli_options_dict()
        opts_string = ' '.join([f"{k}={v}" for k,v in opts.items()])
        startup_files = self._options.get_opa_startup_files()
        return f"opa run --server {opts_string} {startup_files}".strip()

    def _terminate_opa(self):
        logger.info("Stopping OPA")
        self._process.terminate()

    async def _run(self):
        self._init_events()
        while not self._should_stop.is_set():
            for task in asyncio.as_completed([self._run_opa_until_terminated(), self._should_stop.wait()]):
                await task
                break

    @retry(wait=wait_random_exponential(multiplier=0.5, max=10))
    async def _run_opa_until_terminated(self) -> int:
        """
        This function runs opa server as a subprocess.
        it returns only when the process terminates.
        """
        logger.info("Running OPA inline: {command}", command=self.command)
        self._process = await asyncio.create_subprocess_shell(
            self.command,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            start_new_session=True,
        )

        # waits until the process is up, then runs a callback
        asyncio.create_task(wait_until_process_is_up(self._process.pid, callback=self._run_start_callbacks))

        await asyncio.wait([
            pipe_opa_logs(self._process.stdout),
            pipe_opa_logs(self._process.stderr)
        ])

        return_code = await self._process.wait()
        logger.info("OPA exited with return code: {return_code}", return_code=return_code)
        if return_code > 0: # exception in running opa
            raise Exception(f"OPA exited with return code: {return_code}")
        return return_code

    def register_opa_initial_start_callbacks(self, callbacks: List[AsyncCallback]):
        """
        register a callback to run when OPA is started the first time.
        """
        self._on_opa_initial_start_callbacks.extend(callbacks)

    def register_opa_restart_callbacks(self, callbacks: List[AsyncCallback]):
        """
        register a callback to run when OPA is restarted (i.e: OPA was already up,
        then got terminated, and now is up again). this is most often used to keep
        OPA's cache (policy and data) up-to-date, since OPA is started without policy
        or data. With empty cache, OPA cannot evaluate authorization queries correctly.
        """
        self._on_opa_restart_callbacks.extend(callbacks)

    async def _run_start_callbacks(self):
        """
        runs callbacks after OPA process starts
        """
        # TODO: make policy store expose the /health api of OPA
        await asyncio.sleep(1)

        if self._process_was_never_up_before:
            # no need to rehydrate the first time
            self._process_was_never_up_before = False
            logger.info("Running OPA initial start callbacks")
            asyncio.create_task(self._run_callbacks(self._on_opa_initial_start_callbacks))
        else:
            logger.info("Running OPA rehydration callbacks")
            asyncio.create_task(self._run_callbacks(self._on_opa_restart_callbacks))

    async def _run_callbacks(self, callbacks: List[AsyncCallback]):
        return await asyncio.gather(*(callback() for callback in callbacks))

    def _init_events(self):
        if self._should_stop is None:
            self._should_stop = asyncio.Event()

    @staticmethod
    def setup_opa_runner(
        options: Optional[OpaServerOptions] = None,
        initial_start_callbacks: Optional[List[AsyncCallback]] = None,
        rehydration_callbacks: Optional[List[AsyncCallback]] = None
    ):
        """
        factory for OpaRunner, accept optional callbacks to run in certain lifecycle events.

        Initial Start Callbacks:
            The first time we start opa, we might want to do certain actions (like launch tasks)
            that are dependant on the policy store being up (such as PolicyUpdater, DataUpdater).

        Rehydration Callbacks:
            when opa restarts, its cache is clean and it does not have the state necessary
            to handle authorization queries. therefore it is necessary that we rehydrate the
            cache with fresh state fetched from the server.
        """
        opa_runner = OpaRunner(options=options)
        if initial_start_callbacks:
            opa_runner.register_opa_initial_start_callbacks(initial_start_callbacks)
        if rehydration_callbacks:
            opa_runner.register_opa_restart_callbacks(rehydration_callbacks)
        return opa_runner


from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig
import aiohttp
import json

from typing import List, Optional

from opal_common.http import is_http_error_response
from opal_common.schemas.data import DataUpdateReport
from opal_common.logger import logger
from opal_client.data.fetcher import DataFetcher
from opal_client.callbacks.register import CallbackConfig, CallbacksRegister

class CallbacksReporter:
    """
    can send a report to callbacks registered on the callback register
    """
    def __init__(self, register: CallbacksRegister, data_fetcher: DataFetcher) -> None:
        self._register = register
        self._fetcher = data_fetcher

    async def report_update_results(self, report: DataUpdateReport, extra_callbacks: Optional[List[CallbackConfig]] = None):
        try:
            # all the urls that will be eventually called by the fetcher
            urls = []
            report_data = report.json()

            # first we add the callback urls from the callback register
            for entry in self._register.all():
                config = entry.config or HttpFetcherConfig() # should not be None if we got it from the register
                config.data = report_data
                urls.append((entry.url, config))

            # next we add the "one time" callbacks from extra_callbacks (i.e: callbacks sent as part of a DataUpdate message)
            if extra_callbacks is not None:
                for url, config in extra_callbacks:
                    config.data = report_data
                    urls.append((url, config))

            logger.info("Reporting the update to requested callbacks", urls=repr(urls))
            report_results = await self._fetcher.handle_urls(urls)
            # log reports which we failed to send
            for (url, config, result) in report_results:
                if isinstance(result, Exception):
                    logger.error("Failed to send report to {url}, info={exc_info}", url=url, exc_info=repr(result))
                if isinstance(result, aiohttp.ClientResponse) and is_http_error_response(result): # error responses
                    try:
                        error_content = await result.json()
                    except json.JSONDecodeError:
                        error_content = await result.text()
                    logger.error(
                        "Failed to send report to {url}, got response code {status} with error: {error}",
                        url=url,
                        status=result.status,
                        error=error_content
                    )
        except:
            logger.exception("Failed to excute report_update_results")


from typing import List
from fastapi import APIRouter, Depends, status, HTTPException, Response
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR

from opal_common.logger import logger
from opal_common.schemas.security import PeerType
from opal_common.schemas.data import CallbackEntry
from opal_common.authentication.deps import JWTAuthenticator
from opal_common.authentication.verifier import Unauthorized
from opal_common.authentication.types import JWTClaims
from opal_common.authentication.authz import require_peer_type

from opal_client.config import opal_client_config
from opal_client.callbacks.register import CallbacksRegister


def init_callbacks_api(authenticator: JWTAuthenticator, register: CallbacksRegister):
    async def require_listener_token(claims: JWTClaims = Depends(authenticator)):
        try:
            require_peer_type(authenticator, claims, PeerType.listener) # may throw Unauthorized
        except Unauthorized as e:
            logger.error(f"Unauthorized to publish update: {repr(e)}")
            raise

    # all the methods in this router requires a valid JWT token with peer_type == listener
    router = APIRouter(prefix="/callbacks", dependencies=[Depends(require_listener_token)])

    @router.get("", response_model=List[CallbackEntry])
    async def list_callbacks():
        """
        list all the callbacks currently registered by OPAL client.
        """
        return list(register.all())

    @router.get("/{key}", response_model=CallbackEntry)
    async def get_callback_by_key(key: str):
        """
        get a callback by its key (if such callback is indeed registered).
        """
        callback = register.get(key)
        if callback is None:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="no callback found with this key")
        return callback

    @router.post("", response_model=CallbackEntry)
    async def register_callback(entry: CallbackEntry):
        """
        register a new callback by OPAL client, to be called on OPA state updates.
        """
        saved_key = register.put(url=entry.url, config=entry.config, key=entry.key)
        saved_entry = register.get(saved_key)
        if saved_entry is None:
            raise HTTPException(status_code=HTTP_500_INTERNAL_SERVER_ERROR, detail="could not register callback")
        return saved_entry

    @router.delete("/{key}", status_code=status.HTTP_204_NO_CONTENT)
    async def get_callback_by_key(key: str):
        """
        unregisters a callback identified by its key (if such callback is indeed registered).
        """
        callback = register.get(key)
        if callback is None:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail="no callback found with this key")
        register.remove(key)
        return Response(status_code=status.HTTP_204_NO_CONTENT)

    return router



import aiohttp

from fastapi import Response


async def proxy_response(response: aiohttp.ClientResponse) -> Response:
    content = await response.text()
    return Response(
        content=content,
        status_code=response.status,
        headers=dict(response.headers),
        media_type="application/json",
    )




import hashlib
from typing import Dict, Tuple, List, Union, Optional, Generator

from opal_common.logger import logger
from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig
from opal_common.schemas.data import CallbackEntry
from opal_client.config import opal_client_config


CallbackConfig = Tuple[str, HttpFetcherConfig]


class CallbacksRegister:
    """
    A store for callbacks to other services, invoked on OPA state changes.

    Every time OPAL client successfully finishes a transaction to update OPA state,
    all the callbacks in this register will be called.
    """
    def __init__(self, initial_callbacks: Optional[List[Union[str, CallbackConfig]]] = None) -> None:
        self._callbacks: Dict[str, CallbackConfig] = {}
        if initial_callbacks is not None:
            self._load_initial_callbacks(initial_callbacks)
        logger.info("Callbacks register loaded")

    def _load_initial_callbacks(self, initial_callbacks: List[Union[str, CallbackConfig]]) -> None:
        normalized_callbacks = self.normalize_callbacks(initial_callbacks)
        for callback in normalized_callbacks:
            url, config = callback
            key = self.calc_hash(url, config)
            self._register(key, url, config)

    def normalize_callbacks(self, callbacks: List[Union[str, CallbackConfig]]) -> List[CallbackConfig]:
        normalized_callbacks = []
        for callback in callbacks:
            if isinstance(callback, str):
                url = callback
                config = opal_client_config.DEFAULT_UPDATE_CALLBACK_CONFIG
                normalized_callbacks.append((url, config))
            elif isinstance(callback, CallbackConfig):
                normalized_callbacks.append(callback)
            else:
                logger.warning(f"Unsupported type for callback config: {type(callback).__name__}")
                continue
        return normalized_callbacks

    def _register(self, key: str, url: str, config: HttpFetcherConfig):
        self._callbacks[key] = (url, config)

    def calc_hash(self, url: str, config: HttpFetcherConfig) -> str:
        """
        gets a unique hash key from a callback url and config.
        """
        m = hashlib.sha256()
        m.update(url.encode())
        m.update(config.json().encode())
        return m.hexdigest()

    def get(self, key: str) -> Optional[CallbackEntry]:
        """
        gets a registered callback by its key, or None if no such key found in register.
        """
        callback = self._callbacks.get(key, None)
        if callback is None:
            return None
        (url, config) = callback
        return CallbackEntry(key=key, url=url, config=config)

    def put(self, url: str, config: Optional[HttpFetcherConfig] = None, key: Optional[str] = None) -> str:
        """
        puts a callback in the register.
        if no config is provided, the default callback config will be used.
        if no key is provided, the key will be calculated by hashing the url and config.
        """
        default_config = opal_client_config.DEFAULT_UPDATE_CALLBACK_CONFIG
        if isinstance(default_config, dict):
            default_config = HttpFetcherConfig(**default_config)

        callback_config = config or default_config
        auto_key = self.calc_hash(url, callback_config)
        callback_key = key or auto_key
        # if the same callback is already registered with another key - remove that callback.
        # there is no point in calling the same callback twice.
        self.remove(auto_key)
        # register the callback under the intended key (auto-generated or provided)
        self._register(callback_key, url, callback_config)
        return callback_key

    def remove(self, key: str):
        """
        removes a callback from the register, if exists.
        """
        if key in self._callbacks:
            del self._callbacks[key]

    def all(self) -> Generator[CallbackEntry, None, None]:
        """
        a generator yielding all the callback configs currently registered.

        Yields:
            the next callback config found
        """
        for key, (url, config) in iter(self._callbacks.items()):
            yield CallbackEntry(key=key, url=url, config=config)


from typing import Optional
from fastapi import APIRouter, status, HTTPException

from opal_common.logger import logger
from opal_client.data.updater import DataUpdater

def init_data_router(data_updater: Optional[DataUpdater]):
    router = APIRouter()

    @router.post("/data-updater/trigger", status_code=status.HTTP_200_OK)
    async def trigger_policy_data_update():
        logger.info("triggered policy data update from api")
        if data_updater:
            await data_updater.get_base_policy_data(data_fetch_reason="request from sdk")
            return {"status": "ok"}
        else:
            raise HTTPException(
                status_code=status.HTTP_503_SERVICE_UNAVAILABLE,
                detail="Data Updater is currently disabled. Dynamic data updates are not available.")

    return router



from enum import Enum
from typing import Optional, List
from pydantic import BaseModel, Field

class LogLevel(str, Enum):
    info = 'info'
    debug = 'debug'
    error = 'error'

class AuthenticationScheme(str, Enum):
    off = 'off'
    token = 'token'
    tls = 'tls'

class AuthorizationScheme(str, Enum):
    off = 'off'
    basic = 'basic'

class OpaServerOptions(BaseModel):
    """
    Options to configure OPA server (apply when choosing to run OPA inline).

    Security options are explained here in detail: https://www.openpolicyagent.org/docs/latest/security/
    these include:
    - addr (use https:// to apply TLS on OPA server)
    - authentication (affects how clients are authenticating to OPA server)
    - authorization (toggles the data.system.authz.allow document as the authz policy applied on each request)
    - tls_ca_cert_file (CA cert for the CA signing on *client* tokens, when authentication=tls is on)
    - tls_cert_file (TLS cert for the OPA server HTTPS)
    - tls_private_key_file (TLS private key for the OPA server HTTPS)
    """
    addr: str = Field(":8181", description="listening address of the opa server (e.g., [ip]:<port> for TCP)")
    authentication: AuthenticationScheme = Field(AuthenticationScheme.off, description="opa authentication scheme (default off)")
    authorization: AuthorizationScheme = Field(AuthorizationScheme.off, description="opa authorization scheme (default off)")
    config_file: Optional[str] = Field(None, description="path of opa configuration file (format defined here: https://www.openpolicyagent.org/docs/latest/configuration/)")
    tls_ca_cert_file: Optional[str] = Field(None, description="path of TLS CA cert file")
    tls_cert_file: Optional[str] = Field(None, description="path of TLS certificate file")
    tls_private_key_file: Optional[str] = Field(None, description="path of TLS private key file")
    log_level: LogLevel = Field(LogLevel.info, description="log level for opa logs")
    files: Optional[List[str]] = Field(None, description="list of built-in rego policies and data.json files that must be loaded into OPA on startup. e.g: system.authz policy when using --authorization=basic, see: https://www.openpolicyagent.org/docs/latest/security/#authentication-and-authorization")

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True

        @classmethod
        def alias_generator(cls, string: str) -> str:
            """
            converts field named tls_private_key_file to --tls-private-key-file (to be used by opa cli)
            """
            return "--{}".format(string.replace('_', '-'))

    def get_cli_options_dict(self):
        """
        returns a dict that can be passed to the OPA cli
        """
        return self.dict(exclude_none=True, by_alias=True, exclude={'files'})

    def get_opa_startup_files(self) -> str:
        """
        returns a list of startup policies and data
        """
        files = self.files if self.files is not None else []
        return " ".join(files)


from typing import List
from pathlib import Path

from opal_common.paths import PathUtils
from opal_client.config import opal_client_config

def default_subscribed_policy_directories() -> List[str]:
    """
    wraps the configured value of POLICY_SUBSCRIPTION_DIRS, but dedups intersecting dirs.
    """
    subscription_directories = [Path(d) for d in opal_client_config.POLICY_SUBSCRIPTION_DIRS]
    non_intersecting_directories = PathUtils.non_intersecting_directories(subscription_directories)
    return [str(directory) for directory in non_intersecting_directories]


from fastapi_websocket_pubsub.rpc_event_methods import RpcEventClientMethods

from opal_client.logger import logger


class TenantAwareRpcEventClientMethods(RpcEventClientMethods):
    """
    use this methods class when the server uses `TenantAwareRpcEventServerMethods`.
    """
    TOPIC_SEPARATOR = "::"

    async def notify(self, subscription=None, data=None):
        topic = subscription["topic"]
        logger.info("Received notification of event: {topic}", topic=topic, subscription=subscription, data=data)
        if self.TOPIC_SEPARATOR in topic:
            topic_parts = topic.split(self.TOPIC_SEPARATOR)
            if len(topic_parts) > 1:
                topic = topic_parts[1] # index 0 holds the app id
        await self.client.trigger_topic(topic=topic, data=data)
